#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`649`-compliant **annotations** (i.e., ``__annotation__``
dunder dictionaries under Python >= 3.14 dynamically created by
``__annotate__()`` dunder methods, mapping from the names of annotated child
objects of parent hintables to the type hints annotating those child objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: Also, don't neglect to *IMMEDIATELY* excise the
#@method_cached_arg_by_id decorator. Quite a facepalm there, folks.

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep649Exception
from beartype.typing import Optional
from beartype._cave._cavefast import Format  # pyright: ignore
from beartype._data.typing.datatyping import (
    Pep649Hintable,
    Pep649HintableAnnotations,
    TypeException,
)
# from beartype._util.kind.maplike.utilmapfrozen import FrozenDict
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_14
from beartype._util.text.utiltextlabel import label_object

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_pep649_hintable_annotations(
    # Mandatory parameters.
    hintable: Pep649Hintable,

    # Optional parameters.
    hint_format: Format = Format.FORWARDREF,
    exception_cls: TypeException = BeartypeDecorHintPep649Exception,
    exception_prefix: str = '',
) -> Pep649HintableAnnotations:
    '''
    **Immutable memoized annotations** (i.e., :class:`.FrozenDict` instance
    encapsulating the possibly empty ``__annotations__`` dunder dictionary
    mapping from the name of each annotated child object of the passed hintable
    to the type hint annotating that child object) annotating the passed
    **hintable** (i.e., ideally pure-Python object defining the
    ``__annotations__`` dunder dictionary as well as the :pep:`649`-compliant
    ``__annotate__()`` dunder method if the active Python interpreter targets
    Python >= 3.14) if this hintable defines the ``__annotations__`` dunder
    dictionary *or* raise an exception otherwise (i.e., if this hintable fails
    to define the ``__annotations__`` dunder dictionary).

    This getter is memoized for efficiency, guaranteeing amortized worst-case
    :math:`O(1)` constant time complexity. The first call to this getter passed
    a new hintable annotated by one or more type hints containing :math:`n`
    unquoted forward references exhibits non-amortized worst-case :math:`O(n)`
    linear time complexity, justifying the memoization of this getter.

    Parameters
    ----------
    hintable : Pep649Hintable
        Hintable to be inspected.
    hint_format : Format, default: Format.FORWARDREF
        Format of annotated hints to be returned. Defaults to
        :attr:`Format.FORWARDREF`, in which case this getter safely encapsulates
        each otherwise unsafe unquoted forward reference transitively
        subscripting each hint annotating this hintable with a safe
        :class:`annotationlib.ForwardRef` object. Note that the remaining
        formats are situational at best. Specifically:

        * The :attr:`Format.VALUE` format is useful *only* to detect whether
          this hintable is annotated by one or more unquoted forward references
          or not. These hintables occasionally require special-case handling
          elsewhere, which this format facilitates. Notably, if this hintable is
          annotated by one or more unquoted forward references, this getter
          raises a :exc:`NameError` exception when passed this format.
        * The :attr:`Format.STRING` format is useful mostly just for
          documentation purposes. A tangential use case does *occasionally*
          arise, though: comparing annotations dictionaries of two hintables
          annotated by one or more unquoted forward references such that one of
          those dictionaries was postponed under :pep:`563` (i.e., ``from
          __future__ import annotations``). These dictionaries are comparable
          under this format but *not* the default :attr:`Format.FORWARDREF`
          format. Why? **Forward reference proxies** (i.e.,
          :class:`annotationlib.ForwardRef` objects). Whereas
          :attr:`Format.FORWARDREF` injects incomparable forward reference
          proxies into these dictionaries that effectively prohibit dictionary
          comparisons, this format just preserves unquoted forward references
          in the strings it returns.
    exception_cls : TypeException, default: BeartypeDecorHintPep649Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep649Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Pep649HintableAnnotations
        ``__annotations__`` dunder dictionary set on this hintable.

    Raises
    ------
    exception_cls
         If this hintable fails to define the ``__annotations__`` dunder
         dictionary. Since *all* pure-Python hintables (including unannotated
         hintables) define this dictionary, this getter raises an exception only
         if the passed hintable is either:

         * *Not* actually a hintable.
         * A **pseudo-callable object** (i.e., otherwise uncallable object whose
           class renders all instances of that class callable by defining the
           ``__call__()`` dunder method).
    '''

    # "__annotations__" dictionary dictionary defined by this hintable if this
    # hintable is actually a hintable *OR* "None" otherwise.
    hint_annotations = get_pep649_hintable_annotations_or_none(
        hintable=hintable,
        hint_format=hint_format,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If this hintable is *NOT* actually a hintable, raise an exception.
    if hint_annotations is None:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not class.')
        assert issubclass(exception_cls, Exception), (
            f'{repr(exception_cls)} not exception subclass.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise a human-readable exception.
        raise exception_cls(
            f'{exception_prefix}{label_object(hintable)} '
            f'not annotatable by type hints '
            f'(i.e., PEP 649 "__annotate__" and "__annotations__" '
            f'dunder attributes undefined).'
        )
    # Else, that hintable is a hintable.

    # Return this dictionary.
    return hint_annotations

# ....................{ VERSIONS                           }....................
# If the active Python interpreter targets Python >= 3.14...
if IS_PYTHON_AT_LEAST_3_14:
    # ....................{ IMPORTS                        }....................
    # Defer version-specific imports.
    from annotationlib import get_annotations  # type: ignore[import-not-found]
    from beartype._data.kind.datakindiota import SENTINEL
    from beartype._util.error.utilerrget import get_name_error_attr_name
    from beartype._util.cache.utilcacheobjattr import (
        ObjectAttrTypes,
        OBJECT_ATTR_CACHE_LOCK,
        get_object_attr_cached_or_sentinel,
        set_object_attr_cached,
    )

    # ....................{ GETTERS                        }....................
    #FIXME: Unit test us up, please.
    def get_pep649_hintable_annotations_or_none(  # pyright: ignore
        # Mandatory parameters.
        hintable: Pep649Hintable,

        # Optional parameters.
        hint_format: Format = Format.FORWARDREF,
        exception_cls: TypeException = BeartypeDecorHintPep649Exception,
        exception_prefix: str = '',
    ) -> Optional[Pep649HintableAnnotations]:

        # ....................{ PEP 649                    }....................
        # If the caller requested the default "FORWARDREF" format...
        if hint_format is Format.FORWARDREF:
            # For efficiency, attempt to first assume that this hintable's
            # "__annotations__" dunder dictionary complies with the non-default
            # "VALUE" format (i.e., if this hintable is annotated by type hints
            # transitively subscripted by *NO* unquoted forward references).
            # Unquoted forward references are expected to be reasonably rare.
            # So, this is the common case and thus a helpful optimization goal.
            try:
                return getattr(hintable, '__annotations__', None)
            # If this hintable's __annotate__() dunder method underlying its
            # "__annotations__" dunder dictionary raised a "NameError" exception
            # when passed the non-default "VALUE" format by CPython, this
            # hintable does *NOT* comply with the format and is thus annotated
            # by type hints transitively subscripted by one or more unquoted
            # forward references. In this case, fallback to support unquoted
            # forward references via full-blown memoized "FORWARDREF" handling.
            except NameError:
                # If this hintable is a pure-Python function, type, or module,
                # this hintable supports caching of arbitrary attributes. In
                # this case...
                if isinstance(hintable, ObjectAttrTypes):
                    # "__annotations__" dunder dictionary in the "FORWARDREF"
                    # format previously cached by a prior call to either this or
                    # the set_pep649_hintable_annotations() function if any *OR*
                    # the sentinel placeholder otherwise.
                    hintable_annotations = get_object_attr_cached_or_sentinel(
                        obj=hintable,
                        attr_name_if_obj_function=(
                            _ANNOTATIONS_ATTR_NAME_IF_OBJ_FUNC),
                        attr_name_if_obj_type_or_module=(
                            _ANNOTATIONS_ATTR_NAME_IF_OBJ_TYPE_OR_MODULE),
                    )

                    # If such a dictionary was cached, return this dictionary.
                    if hintable_annotations is not SENTINEL:
                        # print(f'Returning hintable {hintable} annotations {hintable_annotations}...')
                        return hintable_annotations  # type: ignore[return-value]
                    # Else, *NO* such dictionary was cached. In this case...
                    else:
                        # print(f'Caching hintable {hintable} annotations...')

                        # Dictionary in this format.
                        #
                        # Note that this dictionary is guaranteed to both exist
                        # (i.e., be non-"None") and non-empty (i.e., contain one
                        # or more hints). Why? Because accessing the
                        # "__annotations__" dunder dictionary raised a
                        # "NameError" exception above, implying this dictionary
                        # to not only exist but contain one or more hints
                        # transitively subscripted by one or more unquoted
                        # forward references.
                        hintable_annotations = (
                            _get_pep649_hintable_annotations_or_none_uncached(  # type: ignore[assignment]
                                hintable=hintable,
                                hint_format=hint_format,
                                exception_cls=exception_cls,
                                exception_prefix=exception_prefix,
                            ))

                        # Cache this dictionary with this hintable for efficient
                        # lookup by a subsequent call to this getter.
                        set_object_attr_cached(
                            obj=hintable,
                            attr_name_if_obj_function=(
                                _ANNOTATIONS_ATTR_NAME_IF_OBJ_FUNC),
                            attr_name_if_obj_type_or_module=(
                                _ANNOTATIONS_ATTR_NAME_IF_OBJ_TYPE_OR_MODULE),
                            attr_value=hintable_annotations,
                        )
                    # Else, such a dictionary was memoized.
                # Else, this hintable is neither a pure-Python function, type,
                # *NOR* module, implying this hintable to *NOT* support caching
                # of arbitrary attributes.
        # Else, the caller did *NOT* request the default "FORWARDREF" format.

        # ....................{ PEP 484                    }....................
        # If the caller requested the non-default "VALUE" format, trivially
        # return this hintable's existing "__annotations__" dunder dictionary.
        # If this is the first access of this dictionary, doing so implicitly:
        # 1. Invokes this hintable's __annotate__() dunder method with this same
        #    non-default "VALUE" format.
        # 2. Caches the returned "__annotations__" dunder dictionary inside this
        #    hintable. This getter avoids re-caching this dictionary.
        elif hint_format is Format.VALUE:
            return getattr(hintable, '__annotations__', None)
        # Else, the caller requested another non-default format (e.g.,
        # "STRING"). Since this format is so situational as to be functionally
        # useless for most intents and purposes, this getter avoids caching the
        # "__annotations__" dunder dictionary unique to this format altogether.

        # ....................{ FALLBACK                   }....................
        # Fallback to the unmemoized getter underlying this memoized getter.
        # Although non-ideal, the only general-purpose alternative would be to
        # memoize a reference to this object, preventing this object from *EVER*
        # being garbage-collected, inviting memory leaks. In other words, there
        # exist *NO* safe means of memoizing arbitrary user-defined objects.
        return _get_pep649_hintable_annotations_or_none_uncached(
            hintable=hintable,
            hint_format=hint_format,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )

    # ....................{ SETTERS                        }....................
    #FIXME: Detect and handle the (possibly common) edge case in which this
    #setter has already been passed this hintable at least once, in which case
    #this hintable's __annotate__() dunder method has *ALREADY* been replaced
    #with an __annotate_beartype__() monkey-patch. We *DEFINITELY* shouldn't
    #keep doing that ad nauseum. Instead, any existing __annotate_beartype__()
    #monkey-patch should be replaced inline *WITHOUT* deferring to
    #"hintable_annotate_old", which in this case will be that existing
    #__annotate_beartype__(). Basically, we just need to generalize this:
    #    hintable_annotate_old = getattr(hintable, '__annotate__', None)
    #
    #...to additionally check whether that hintable is already an
    #__annotate_beartype__() monkey-patch and, if so, defer to the *TRUE*
    #original __annotate__() dunder method. What's the catch? We don't currently
    #preserve the original __annotate__() dunder method in a @beartype-specific
    #dunder attribute on our __annotate_beartype__() monkey-patches. I suppose
    #we'll need to start doing that. The logic resembles:
    #    hintable_annotate_old = getattr(hintable, '__annotate__', None)
    #    hintable_annotate_old_wrappee = getattr(
    #        hintable_annotate_old, '__beartype_annotate_wrappee__', None)
    #    if hintable_annotate_old_wrappee is not None:
    #        hintable_annotate_old = hintable_annotate_old_wrappee
    #
    #Then we'll just need to monkey-patch the "__beartype_annotate_wrappee__"
    #attribute into our __annotate_beartype__() monkey-patch. *sigh*
    #FIXME: Oh. Right. The above is *BASICALLY* almost perfect, except that
    #there's no need for a @beartype-specific "__beartype_annotate_wrappee__"
    #monkey-patch. Instead, just use @functools.wraps(hintable_annotate_old) as
    #Guido intended. High fives all around, Team Bear! \o/

    #FIXME: Unit test us up, please.
    def set_pep649_hintable_annotations(
        # Mandatory parameters.
        hintable: Pep649Hintable,
        annotations: Pep649HintableAnnotations,

        # Optional parameters.
        exception_cls: TypeException = BeartypeDecorHintPep649Exception,
        exception_prefix: str = '',
    ) -> None:
        assert isinstance(annotations, dict), (
            f'{repr(annotations)} not dictionary.')
        assert all(
            isinstance(annotations_key, str) for annotations_key in annotations
        ), f'{repr(annotations)} not dictionary mapping names to type hints.'
        # print(f'Setting hintable {hintable} annotations to {annotations}...')

        # ....................{ PREAMBLE                   }....................
        # Thread-safely...
        with OBJECT_ATTR_CACHE_LOCK:
            # ....................{ CACHE                  }....................
            # If this hintable is *NOT* actually a hintable, raise an exception.
            # Amusingly, the simplest means of implementing this validation is
            # to simply retrieve the existing "__annotations__" dunder
            # dictionary currently defined on this hintable.
            get_pep649_hintable_annotations(
                hintable=hintable,
                exception_cls=exception_cls,
                exception_prefix=exception_prefix,
            )
            # Else, this hintable is actually a hintable defining the requisite
            # pair of PEP 649- and 749-compliant dunder attributes:
            # * __annotate__().
            # * "__annotations__".

            # If this hintable is a pure-Python function, type, or module, this
            # hintable supports caching of arbitrary attributes. In this case,
            # re-cache this dictionary onto this hintable to avoid
            # desynchronization with a prior "__annotations__" dunder dictionary
            # cached previously onto this hintable.
            if isinstance(hintable, ObjectAttrTypes):
                set_object_attr_cached(
                    obj=hintable,
                    attr_name_if_obj_function=(
                        _ANNOTATIONS_ATTR_NAME_IF_OBJ_FUNC),
                    attr_name_if_obj_type_or_module=(
                        _ANNOTATIONS_ATTR_NAME_IF_OBJ_TYPE_OR_MODULE),
                    attr_value=annotations,
                )
            # Else, this hintable is neither a pure-Python function, type, *NOR*
            # module, implying this hintable does *NOT* support caching of
            # arbitrary attributes.

            # ....................{ LOCALS                 }....................
            # Existing __annotate__() dunder method set on this hintable if any
            # *OR* "None" (e.g., if an external caller has already explicitly
            # set the "__annotations__" dunder attribute on this hintable, which
            # implicitly sets the __annotate__() dunder method to "None").
            hintable_annotate_old = getattr(hintable, '__annotate__', None)

            # Either:
            # * If this hintable is annotated by type hints transitively
            #   subscripted by one or more unquoted forward references, the
            #   "NameError" exception implicitly raised by attempting to access
            #   the existing "__annotations__" dunder dictionary set on this
            #   hintable cached according to the "VALUE" format.
            # * If this hintable is annotated by type hints transitively
            #   subscripted by *NO* unquoted forward references, "None".
            hintable_annotations_old_name_error: Optional[Exception] = None

            # Attempt to...
            try:
                # Existing "__annotations__" dunder dictionary set on this
                # hintable cached according to the non-default "VALUE" format if
                # this hintable is annotated by type hints transitively
                # subscripted by *NO* unquoted forward references *OR*
                # implicitly raise the "NameError" exception otherwise (i.e., if
                # this hintable is annotated by type hints transitively
                # subscripted by one or more unquoted forward references).
                hintable.__annotations__
            # If accessing this dictionary above raised a "NameError" exception,
            # this hintable is annotated by type hints transitively subscripted
            # by one or more unquoted forward references. Preserve this
            # exception for subsequent re-raising below.
            except NameError as exception:
                hintable_annotations_old_name_error = exception

            # ....................{ CLOSURE                }....................
            def __annotate_beartype__(
                hint_format: Format) -> Pep649HintableAnnotations:
                f'''
                Hintable {repr(hintable)} :pep:`649`- and :pep:`749`-compliant
                ``__annotate__()`` dunder method, modifying the user-defined
                ``__annotations__`` dunder dictionary for this hintable with
                :mod:`beartype`-specific improvements.

                These improvements include:

                * **Memoization** (i.e., caching) across root type hints,
                  reducing space consumption.
                * :pep:`563`-compliant conversion of unquoted forward references
                  under the ``from __future__ import annotations`` pragma into
                  equivalent :mod:`beartype`-specific **forward reference
                  proxies** (i.e., objects proxying undefined types).

                This getter dunder method is intentionally given a
                :mod:`beartype`-specific name to aid in external debugging.

                Parameters
                ----------
                hint_format : Format
                    Kind of annotation format to be returned. See also
                    :pep:`649` and :pep:`749` for further details.

                Returns
                -------
                Pep649HintableAnnotations
                    ``__annotations__`` dunder dictionary set on this hintable.
                '''

                #FIXME: Submit an upstream CPython issue about this. This
                #behaviour is super-weird, non-orthogonal, and invites extremely
                #subtle and non-trivial to debug issues in user code like this.
                #To resolve this, CPython devs should consider:
                #* Defining a new "_annotationlib" C extension.
                #* Moving the existing "annotationlib.Format" enum to this C
                #  extension.
                #* Adding to the top of "annotationlib":
                #      from _annotationlib import Format
                #* Refactoring the C-based CPython interpreter to pass the
                #  "_annotationlib.Format.VALUE" enum member rather than the
                #  magic integer constant "1" to __annotate__() dunder methods
                #  when creating the "__annotations__" dunder dictionary.
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                # CAUTION: CPython implicitly calls this monkey-patched dunder
                # function with magic integer constants (e.g., "1") rather than
                # readable enum members (e.g., "Format.FORWARDREF"). In other
                # words, *THIS FUNCTION MUST NOT ATTEMPT TO COMPARE THE PASSED
                # PARAMETER TO ENUM MEMBERS WITH THE "is" BUILTIN.* Doing so is
                # guaranteed to silently fail in non-debuggable ways with
                # sporadic false negatives or positives. We know. We were there.
                #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                #FIXME: [SPEED] Globalize access to frequently accessed "Format"
                #members and reference those globals instead below. This method
                #*COULD* be frequently called enough to warrant micro-optimization.

                # If the caller requested the default "FORWARDREF" format,
                # trivially return the "__annotations__" dunder dictionary
                # passed by the original earlier caller to the parent
                # set_pep649_hintable_annotations() setter of this closure.
                #
                # If this dictionary contains:
                # * *NO* unquoted forward references, this dictionary already
                #   complies with the "VALUE" format. However, the "FORWARDREF"
                #   format simply reduces to "VALUE" if a dictionary contains
                #   *NO* unquoted forward references. Ergo, this dictionary also
                #   complies with the "FORWARDREF" format (which implements the
                #   superset of "VALUE").
                # * One or more unquoted forward references, this dictionary
                #   already complies with the default "FORWARDREF" format. Why?
                #   A complex and non-obvious chain of casuistry. Bear with us.
                #   If this dictionary did *NOT* already comply with the
                #   "FORWARDREF" format, then (by process of elimination)
                #   this dictionary *MUST* at least comply with the "VALUE"
                #   format. However, since this dictionary contains unquoted
                #   forward references, the annotationlib.get_annotations()
                #   getter would have raised a "NameError" exception when
                #   attempting to return this dictionary under the "VALUE"
                #   format, in which case this dictionary could *NOT* possibly
                #   exist. Clearly, however, this dictionary exists. Since this
                #   is an obvious contradiction, this dictionary *MUST*
                #   necessarily already comply with the "FORWARDREF" format.
                #
                # There now exist two possible cases:
                # * The caller passed "FORWARDREF" and this dictionary complies
                #   with the "FORWARDREF" format. *WONDERFUL!*
                # * The caller passed "FORWARDREF" and this dictionary only
                #   complies with the "VALUE" format, which implies this
                #   dictionary contains *NO* unquoted forward references. Since
                #   the "FORWARDREF" format simply reduces to "VALUE" if a
                #   dictionary contains *NO* unquoted forward references, this
                #   case is still *WONDERFUL!*
                if hint_format == Format.FORWARDREF:  # <-- "==", *NOT* "is"!
                    return annotations
                # Else, the caller did *NOT* request the "FORWARDREF" format.
                #
                # If the caller requested the "VALUE" format...
                elif hint_format == Format.VALUE:  # <-- "==", *NOT* "is"!
                    # If attempting to access the existing "__annotations__"
                    # dunder dictionary set on this hintable cached according to
                    # the non-default "VALUE" format raised a "NameError"
                    # exception above, this hintable was annotated by type hints
                    # transitively subscripted by one or more unquoted forward
                    # references. In this case, re-raise the same exception to
                    # notify the caller of this critical fact.
                    if hintable_annotations_old_name_error is not None:
                        raise hintable_annotations_old_name_error
                    # Else, doing so did *NOT* raise a "NameError" exception,
                    # implying this hintable was annotated by type hints
                    # transitively subscripted by *NO* unquoted forward
                    # references, implying the old "__annotations__" dunder
                    # dictionary complies with the "Format.VALUE" format.

                    # Return the new "__annotations__" dunder dictionary. By
                    # definition, this dictionary exists and thus implicitly
                    # complies with the "Format.VALUE" format as well.
                    return annotations
                # Else, the caller did *NOT* request the "VALUE" format.
                #
                # If an existing __annotate__() dunder method was previously
                # defined on this hintable...
                elif hintable_annotate_old is not None:
                    # print(f'Caller requested odd format {hint_format}...')
                    return hintable_annotate_old(hint_format)
                # Else, *NO* existing __annotate__() dunder method was
                # previously defined on this hintable. This beartype-specific
                # implementation of that method *MUST* now either:
                # * Attempt to manually redefine this format. Although feasible,
                #   doing so would be non-trivial and undesirable. See above.
                # * Raise the builtin "NotImplementedError" exception. The
                #   caller is then expected to manually implement this format.
                #   Thankfully, the high-level call_annotate_function() and
                #   get_annotations() functions defined by the standard
                #   "annotationslib" module do just that; they explicitly catch
                #   this exception and respond by implementing the "STRING"
                #   format in the expected way. This is the only sane solution.
                #
                # Unsurprisingly, we opt for the latter approach by simply
                # falling through to the fallback defined below.

                # Notify the caller that this __annotate__() implementation
                # fails to support this format by raising a
                # "NotImplementedError" exception.
                #
                # Note that:
                # * PEP 649 itself encourages user-defined __annotate__()
                #   implementations to raise bare "NotImplementedError"
                #   exceptions lacking messages. Indeed, the
                #   call_annotate_function() and get_annotations() functions
                #   trivially catch these exceptions and ignore associated
                #   messages.
                # * PEP 749 explicitly instructs user-defined __annotate__()
                #   implementations to raise "NotImplementedError" exceptions
                #   when passed the private ".VALUE_WITH_FAKE_GLOBALS" format:
                #       Users who manually write annotate functions should raise
                #       NotImplementedError if the VALUE_WITH_FAKE_GLOBALS
                #       format is requested, so the standard library will not
                #       call the manually written annotate function with “fake
                #       globals”, which could have unpredictable results.
                raise NotImplementedError()

            # ....................{ MONKEY-PATCH           }....................
            # Attempt to...
            try:
                # Silently replace this hintable's existing __annotate__()
                # dunder method with this new beartype-specific monkey-patch
                hintable.__annotate__ = __annotate_beartype__  # type: ignore[union-attr]
                # print(f'new annotations: {annotations}')
                # print(f'{hintable}.__annotate__: {hintable.__annotate__}')
                # print(f'{hintable}.__annotations__: {hintable.__annotations__}')
                # print(f'{hintable}.__annotate__(3): {hintable.__annotate__(Format.FORWARDREF)}')
                # hintable_annotations_cached = get_pep649_hintable_annotations(hintable)
                # print(f'{hintable}.__annotate__(3) [cached]: {hintable_annotations_cached}')
            # If doing so fails with an exception resembling the following, this
            # hintable is *NOT* pure-Python. The canonical example are C-based
            # decorator objects (e.g., class, property, or static method
            # descriptors), whose exception message reads:
            #     AttributeError: 'method' object has no attribute
            #     '__annotate__' and no __dict__ for setting new attributes. Did
            #     you mean: '__getstate__'?
            #
            # C-based decorator objects only define:
            # * A read-only __annotate__() dunder method that proxies an
            #   original writeable __annotate__() dunder method of the
            #   pure-Python callables they originally decorated.
            # * A read-only "__annotations__" dunder attribute that proxies an
            #   original writeable "__annotations__" dunder attribute of the
            #   pure-Python callables they originally decorated.
            #
            # Detecting this edge case is non-trivial and most easily deferred
            # to this late time. While non-ideal, simplicity >>>> idealism here.
            except AttributeError as exception:
                # Lower-level presumably pure-Python callable wrapped by this
                # higher-level presumably C-based decorator object if this
                # decorator object wraps such a callable *OR* "None" otherwise
                # (i.e., if this object does *NOT* wrap such a callable).
                #
                # Note that this callable is intentionally accessed as the
                # "__func__" dunder attribute on this decorator object, as the
                # following standard decorator objects *ALL* wrap their
                # decorated callables with this attribute:
                # * Bound instance method descriptors.
                # * @classmethod-decorated class method descriptors.
                # * @staticmethod-decorated static method descriptors.
                #
                # See also the "beartype._util.func.utilfuncwrap" submodule.
                hintable_func = getattr(hintable, '__func__', None)

                #FIXME: File an upstream CPython issue about this, please. *sigh*
                #FIXME: Remove this edge case *AFTER* some future Python version
                #fully satisfies PEP 749 by implementing this paragraph:
                #    The constructors for classmethod() and staticmethod() currently
                #    copy the __annotations__ attribute from the wrapped object to
                #    the wrapper. They will instead have writable attributes for
                #    __annotate__ and __annotations__. Reading these attributes will
                #    retrieve the corresponding attribute from the underlying
                #    callable and cache it in the wrapper’s __dict__. Writing to
                #    these attributes will directly update the __dict__, without
                #    affecting the wrapped callable.
                #
                #Currently, Python does *NOT* do that. Neither the __annotate__()
                #nor "__annotate__" dunder attributes are settable on @classmethod
                #or @staticmethod descriptors:
                #    class Yum(object):
                #        @classmethod
                #        def guh(cls) -> None: pass
                #
                #    def ugh_annotate(): return {}
                #
                #    yim = Yum()
                #    print(Yum.guh.__annotate__)          # <-- reading this works
                #    Yum.guh.__annotate__ = ugh_annotate  # <-- writing this fails
                #
                #The above example currently raises:
                #    AttributeError: 'method' object has no attribute '__annotate__'
                #    and no __dict__ for setting new attributes. Did you mean:
                #    '__getstate__'?
                #
                #Presumably, Python will start doing that at some point. Once Python
                #does, this issue becomes a non-issue. For the moment, efficiency is
                #irrelevant. We just need this to work for a temporary span of time.
                #FIXME: Once Python resolves this issue, also remove the
                #temporary "if IS_PYTHON_AT_MOST_3_13:" hack from our companion
                #test_resolve_pep563() unit test. *sigh*

                # If...
                if (
                    # This higher-level C-based decorator object wraps a
                    # lower-level pure-Python callable *AND*...
                    hintable_func is not None and
                    # This decorator object does *NOT* wrap itself...
                    #
                    # Note that:
                    # * This edge case should never occur. Indeed, if this
                    #   decorator object is one of the standard decorator
                    #   objects listed above, this edge case is guaranteed to
                    #   *NOT* occur.
                    # * This identity test is a Poor Man's Recursion Guard.
                    #   Clearly, this identity test does *NOT* actually
                    #   constitute a recursion guard. Implementing a "true"
                    #   recursion guard would require tracking a set of all
                    #   previously seen hintables across recursive calls. Since
                    #   it's unclear whether this edge
                    #   case will even arise in practice, it's unclear whether
                    #   the effort is worth investing in a "true" recursion
                    #   guard. The Poor Man's Recursion Guard suffices...
                    hintable_func is not hintable
                ):
                    # Set the "__annotations__" dunder dictionary on this
                    # lower-level pure-Python callable *BEFORE* setting the
                    # __annotate__() dunder method on this callable below. Why?
                    # Because CPython currently propagates *ONLY*
                    # "__annotations__" but not __annotate__() from this
                    # lower-level pure-Python callable on up to this
                    # higher-level C-based decorator object. Do so *BEFORE*
                    # setting __annotate__(), which implicitly nullifies
                    # "__annotations__". (Look. All of this is busted. I sigh.)
                    hintable_func.__annotations__ = annotations

                    # Set the __annotate__() dunder method on this lower-level
                    # pure-Python callable.
                    set_pep649_hintable_annotations(
                        hintable=hintable_func,
                        annotations=annotations,
                        exception_cls=exception_cls,
                        exception_prefix=exception_prefix,
                    )
                # Else, either this higher-level presumably C-based decorator
                # object does not wrap a lower-level presumably pure-Python
                # callable *OR* this decorator object wraps itself. In either
                # case, unwrapping this decorator object would be harmful.
                # Unsurprisingly, avoid doing so.
                else:
                    # If the "__annotations__" dunder attribute of this hintable
                    # is *NOT* a dictionary, this dunder attribute has
                    # *PROBABLY* been nullified to "None", *PROBABLY* due to
                    # another decorator having previously set the __annotate__()
                    # dunder method of the presumably pure-Python callable
                    # underlying this C-based decorator object. Yes, there are a
                    # lot of assumptions *PROBABLY* happening here. The only
                    # remaining means of setting the passed annotations
                    # dictionary on this hintable would be to set the
                    # "__annotations__" dunder attribute to this dictionary.
                    # However, attempting to set the __annotate__() dunder
                    # method raised an "AttributeError"! Ergo, attempting to set
                    # the "__annotations__" dunder dictionary would almost
                    # certainly raise the same exception. Since there exist *NO*
                    # remaining means of setting the passed annotations
                    # dictionary on this hintable, we have *NO* recourse but to
                    # notify the caller of this modern tragedy by raising an
                    # exception.
                    if not isinstance(hintable.__annotations__, dict):
                        raise exception_cls(
                            f'{exception_prefix}{label_object(hintable)} '
                            f'type hints not settable to '
                            f'annotations dictionary {repr(annotations)} '
                            f'(i.e., PEP 649 "__annotate__" and "__annotations__" '
                            f'dunder attributes not settable, but "__annotations__" '
                            f'dunder attribute already set to '
                            f'non-dictionary value {repr(hintable.__annotations__)}).'
                        ) from exception
                    # Else, the "__annotations__" dunder attribute of this
                    # hintable is a dictionary.

                    # For the name of each annotated attribute of this hintable
                    # and the new hint which which to annotate this attribute,
                    # overwrite the prior hint originally annotating this
                    # attribute with this new hint.
                    #
                    # Note that:
                    # * The above assignment is an efficient O(1) operation and
                    #   thus intentionally performed first.
                    # * This iteration-based assignment is an inefficient O(n)
                    #   operation (for "n" the number of attributes annotated on
                    #   this hintable) and thus intentionally performed last.
                    for attr_name, attr_hint in annotations.items():
                        hintable.__annotations__[attr_name] = attr_hint

    # ....................{ PRIVATE ~ globals : str        }....................
    _ANNOTATIONS_ATTR_NAME_IF_OBJ_FUNC = '__beartype_annotations'
    '''
    Unique name of the **memoized function annotations attribute** (i.e.,
    attribute cached for each pure-Python function passed to the
    :func:`beartype.beartype` decorator accepting one or more parameters and/or
    returns annotated by type hints whose attribute value is that function's
    **memoized annotations dictionary** (i.e., dictionary from the name of each
    such parameter or return to the type hint annotating that parameter or
    return as returned by the :func:`.get_pep649_hintable_annotations_or_none`
    getter when passed that function)).
    '''


    _ANNOTATIONS_ATTR_NAME_IF_OBJ_TYPE_OR_MODULE = 'annotations'
    '''
    Unique name of the **memoized type or module annotations attribute** (i.e.,
    attribute cached for each pure-Python type or module passed to the
    :func:`beartype.beartype` decorator defining one or more class or global
    variables annotated by type hints whose attribute value is that type's or
    module's **memoized annotations dictionary** (i.e., dictionary from the name
    of each such variable to the type hint annotating that variable as returned
    by the :func:`.get_pep649_hintable_annotations_or_none` getter when passed
    that type or module)).
    '''

    # ....................{ PRIVATE ~ getters              }....................
    def _get_pep649_hintable_annotations_or_none_uncached(
        hintable: Pep649Hintable,
        hint_format: Format,
        exception_cls: TypeException,
        exception_prefix: str,
    ) -> Optional[Pep649HintableAnnotations]:
        '''
        **Immutable unmemoized annotations** (i.e., :class:`.FrozenDict`
        instance encapsulating the possibly empty ``__annotations__`` dunder
        dictionary mapping from the name of each annotated child object of the
        passed hintable to the type hint annotating that child object)
        annotating the passed **hintable** (i.e., ideally pure-Python object
        defining the ``__annotations__`` dunder attribute as well as the
        :pep:`649`-compliant ``__annotate__`` dunder method if the active Python
        interpreter targets Python >= 3.14) if this hintable defines the
        ``__annotations__`` dunder dictionary *or* :data:`None` otherwise (i.e.,
        if this hintable fails to define the ``__annotations__`` dunder
        dictionary).

        This getter exhibits non-amortized worst-case :math:`O(n)` linear time
        complexity for :math:`n` the total number of unquoted forward references
        across all type hints annotating this hintable.

        Parameters
        ----------
        hintable : Pep649Hintable
            Hintable to be inspected.
        hint_format : Format
            Format of annotated hints to be returned.
        exception_cls : TypeException
            Type of exception to be raised in the event of a fatal error.
        exception_prefix : str
            Human-readable substring prefixing raised exception messages.

        Returns
        -------
        Optional[Pep649HintableAnnotations]
            Either:

            * If this hintable is actually a hintable, the ``__annotations__``
              dunder dictionary set on this hintable.
            * Else, :data:`None`.

        Raises
        ------
        exception_cls
            If:

            * This hintable is annotated by one or more type hints transitively
              subscripted by one or more unquoted forward references.
            * This hintable's __annotate__() dunder method has been nullified
              (i.e., previously set to :data:`None` and thus destroyed).
        '''

        # ....................{ LOCALS                     }....................
        # Annotations dictionary to be returned if this hintable is annotated
        # *OR* "None" otherwise (i.e., if this hintable is unannotated).
        hintable_annotations: Optional[Pep649HintableAnnotations] = None

        # ....................{ FORMAT ~ forwardref        }....................
        # If this format requests that unquoted forward references be wrapped by
        # "annotationlib.ForwardRef" objects, do *NOT* unconditionally call the
        # annotationlib.get_annotations() getter. Why? Because that getter
        # raises unreadable exceptions when passed this format under various
        # common edge cases. Instead...
        if hint_format is Format.FORWARDREF:
            # ....................{ PEP 649                }....................
            # If this hintable defines the PEP 649-compliant __annotate__()
            # dunder method to be anything *OTHER* than "None", this hintable is
            # expected to be annotated by one or more type hints.
            #
            # Note that:
            # * The __annotate__() dunder method is guaranteed to exist *ONLY*
            #   for standard pure-Python hintables. Various other hintables of
            #   interest (e.g., functions exported by the standard "operator"
            #   module) do *NOT* necessarily declare this method. Since this
            #   getter is commonly called in general-purpose contexts where this
            #   guarantee does *NOT* necessarily hold, we intentionally access
            #   this attribute safely albeit somewhat more slowly via getattr().
            # * PEP 649 supports external nullification of the __annotate__()
            #   dunder method (i.e., by setting this dunder attribute to
            #   "None"). Indeed, PEP 649 explicitly requires nullification as a
            #   means of efficiently declaring a hintable to be unannotated:
            #       If an object has no annotations, __annotate__ should be
            #       initialized to None, rather than to a function that returns
            #       an empty dict.
            # * The __annotate__() dunder method and "__annotations__" dunder
            #   dictionary invalidate one another. Setting one nullifies the
            #   other:
            #       * Setting o.__annotate__ to a callable invalidates the
            #         cached annotations dict.
            #       * Setting o.__annotations__ to a legal value automatically
            #         sets o.__annotate__ to None.
            #   Thus, the __annotate__() dunder method being "None" does *NOT*
            #   imply this hintable to be unannotated. The "__annotations__"
            #   dunder attribute may be a non-"None" non-empty dictionary, in
            #   which case this hintable would be annotated. Where annotations
            #   are concerned, there are now multiple sources of objective
            #   truth. This is awful.
            # * The get_annotations() getter called below safely accepts the
            #   "FORWARDREF" format *ONLY* when this hintable defines the
            #   __annotate__() dunder method. If this hintable does *NOT* define
            #   __annotate__() and get_annotations() is passed "FORWARDREF",
            #   then get_annotations() raises either:
            #   * If this hintable at least defines the "__annotations__" dunder
            #     dictionary but this dictionary contains one or more unquoted
            #     forward references, a "NameError" exception.
            #   * Else, a "TypeError" exception.
            #
            #   However, this higher-level getter is designed exactly to avoid
            #   raising these sorts of exceptions! Ergo, get_annotations() is
            #   safely callable only when the __annotate__() dunder method
            #   exists.
            if getattr(hintable, '__annotate__', None) is not None:
                # Defer to the PEP 649-compliant high-level
                # annotationlib.get_annotations() getter internally deferring to
                # the PEP 649-compliant low-level __annotate__() dunder callable
                # rather than the PEP 484-compliant "__annotations__" dunder
                # attribute. Why? Because the latter reduces to calling
                # "get_annotations(hintable, format=Format.VALUE)", which raises
                # a "NameError" exception if the passed hintable is annotated by
                # one or more unquoted forward references. This is unacceptable
                # API design. Yet, this is Python >= 3.14.
                #
                # Note that:
                # * get_annotations() is guaranteed to *NEVER* return "None". If
                #   the __annotate__() dunder method and "__annotations__"
                #   dunder attribute are both "None", then get_annotations()
                #   raises a "TypeError" exception. However, get_annotations()
                #   is the canonical means of retrieving annotations under
                #   Python >= 3.14. Thus, we infer that at most one of but *NOT*
                #   both of __annotate__() and "__annotations__" may be "None".
                # * get_annotations() is guaranteed to *ALWAYS* return a new
                #   dictionary rather than the same value as that of the
                #   "__annotations__" dunder dictionary when the passed format
                #   is "FORWARDREF". This inefficiency is baked into
                #   get_annotations() and thus *CANNOT* be avoided.
                hintable_annotations = get_annotations(
                    hintable, format=hint_format)
                # print(f'Found hintable {hintable} annotations {hintable_annotations} in {hint_format}...')
            # Else, this hintable does *NOT* define __annotate__().

            # ....................{ PEP 484                }....................
            # Return either the PEP 484-compliant "__annotations__" dunder
            # dictionary if this hintable defines this dictionary *OR* "None"
            # (i.e., if this hintable fails to define this dictionary).
            #
            # Note that:
            # * The "__annotations__" dunder attribute is guaranteed to exist
            #   *ONLY* for standard pure-Python hintables. See above.
            # * The "__annotations__" dunder attribute is expected to either:
            #   * If this hintable actually is a hintable, be non-"None". Why?
            #     Because the __annotate__() dunder method was "None". By the
            #     logic given above, it should *NEVER* be the case that both
            #     __annotate__() and "__annotations__" are "None". However,
            #     __annotate__() was "None". It follows that "__annotations__"
            #     should now be non-"None" (and thus a valid dictionary).
            #   * Else, *NOT* exist. Ideally, unhintable objects should *NEVER*
            #     define the "__annotations__" dunder attribute.
            #
            #   Ergo, it follows that this getter returns "None" *ONLY* when
            #   this hintable is *NOT* actually a hintable. Sanity preserved!
            # * The "__annotations__" dunder attribute is *NOT* safely
            #   accessible under Python >= 3.14 in the worst case. If this
            #   dictionary contains one or more type hints subscripted by one or
            #   more unquoted forward references, then directly accessing this
            #   attribute is guaranteed to raise a non-human-readable
            #   "NameError" exception. Consequently, we perform this unsafe
            #   fallback *ONLY* when the __annotate__() dunder method does *NOT*
            #   exist. Although non-ideal, PEP 649 explicitly permits callers to
            #   set this attribute -- presumably as an unsafe means of
            #   preserving backward compatibility. That would be fine, except
            #   that setting this attribute nullifies and thus destroys any
            #   previously set __annotate__() dunder method! Again:
            #       * Setting o.__annotations__ to a legal value automatically
            #         sets o.__annotate__ to None.
            # * The "__annotations__" dunder dictionary and __annotate__()
            #   dunder method are strongly coupled. If one is defined, the other
            #   should be defined. If one is undefined, the other should be
            #   undefined. Ergo, it should *NEVER* be the case that the
            #   __annotate__() dunder method is undefined but the
            #   "__annotations__" dunder dictionary is defined. Ergo, this edge
            #   case should *NEVER* arise. Naturally, this edge case will often
            #   arise. Why? Because nothing prevents third-party packages from
            #   manually defining "__annotations__" dunder dictionaries on
            #   arbitrary objects. Although CPython *COULD* prohibit that (e.g.,
            #   by defining the "object.__annotations__" descriptor to do just
            #   that), CPython currently does *NOT* prohibit that. In fact, no
            #   "object.__annotations__" descriptor exists to even do so.
            else:
                # Attempt to fallback to the PEP 484-compliant "__annotations__"
                # dunder dictionary if this hintable defines this dictionary
                # *OR* "None" otherwise.
                try:
                    hintable_annotations = getattr(
                        hintable, '__annotations__', None)
                # If accessing this dictionary raises an unreadable "NameError"
                # exception, this hintable is annotated by one or more type
                # hints transitively subscripted by one or more unquoted forward
                # references. However, this hintable's __annotate__() dunder
                # method has been nullified (i.e., previously set to "None" and
                # thus destroyed)! In this case, calling the get_annotations()
                # getter called above would simply re-raise the same:
                #     NameError: name 'UndefinedType' is not defined
                #
                # While uncommon, this edge case arises when a some previously
                # applied obsolete PEP 649-noncompliant decorator unsafely set
                # the "__annotations__" dunder dictionary on this hintable,
                # which then implicitly nullified the __annotate__() dunder
                # method. Since this constitutes a fatal issue that the caller
                # should be informed about, raise a more readable exception.
                except NameError as exception:
                    # Name of the currently undefined attribute referred to be
                    # the first unquoted forward reference possibly deeply
                    # nested in the first hint annotating this hintable.
                    hint_ref_name = get_name_error_attr_name(exception)

                    # Raise an exception embedding this name.
                    raise exception_cls(
                        f'{exception_prefix}{label_object(hintable)} '
                        f'unsafely annotated by unresolvable type hints, as:\n'
                        f'* One or more type hints transitively subscripted by '
                        f'unquoted forward reference "{hint_ref_name}".\n'
                        f'* __annotate__() dunder method nullified '
                        f'(i.e., previously set to "None" and thus destroyed).\n'
                        f'{repr(hintable)} is presumably decorated by a '
                        f'PEP 649-noncompliant decorator unsafely setting the '
                        f'"__annotations__" dunder attribute, '
                        f'which no decorators should do under Python >= 3.14. '
                        f'Consider submitting an upstream issue report to '
                        f'the authors of that decorator. Politely request that '
                        f'they join the modern world and support PEP 649.'
                    ) from exception
        # ....................{ FORMAT ~ other             }....................
        # Else, this is any format *EXCEPT* the format requesting that unquoted
        # forward references be wrapped by "annotationlib.ForwardRef" objects.
        # In this case, unconditionally call the annotationlib.get_annotations()
        # getter, which does *NOT* raise unexpected exceptions and is thus
        # safely callable when passed this format.
        else:
            hintable_annotations = get_annotations(hintable, format=hint_format)

        # ....................{ RETURN                     }....................
        #FIXME: Actually, let's just return this mutable annotations dictionary
        #as is for the moment. Although non-ideal, this is mostly fine. Why?
        #Because when "hint_format" is the default "Format.FORWARDREF" (which is
        #the case for 99.99% of all calls to this getter), this annotations
        #dictionary is guaranteed to be a copy of the underlying
        #"__annotations__" dunder dictionary. Mutating a copy is always fine. Of
        #course, we then memoize this copy. Ordinarily, mutating a memoized
        #object would absolutely *NOT* be fine. In this case, though, mutating
        #this memoized object is actually ideal. Why? Because then we only need
        #to coerce hints once (e.g., via a call to the coerce_func_hint_root()
        #function), because the result of doing so is then memoized.

        # Return this annotations dictionary, coerced into an immutable frozen
        # dictionary for safety (e.g., to prevent accidental external mutation).
        # return FrozenDict(hintable_annotations)
        return hintable_annotations
# Else, the active Python interpreter targets Python <= 3.13. In this case,
# trivially defer to the PEP 484-compliant "__annotations__" dunder attribute.
else:
    # ....................{ GETTERS                        }....................
    def get_pep649_hintable_annotations_or_none(  # type: ignore[misc]
        hintable: Pep649Hintable, **kwargs) -> (
        Optional[Pep649HintableAnnotations]):

        # Return either the PEP 484-compliant "__annotations__" dunder attribute
        # if the passed hintable defines this attribute *OR* "None" otherwise
        # (i.e., if this hintable fails to define this attribute).
        #
        # Note that the "__annotations__" dunder attribute is guaranteed to
        # exist *ONLY* for standard pure-Python hintables. Various other
        # hintables of interest (e.g., functions exported by the standard
        # "operator" module) do *NOT* necessarily declare this attribute. Since
        # this getter is commonly called in general-purpose contexts where this
        # guarantee does *NOT* necessarily hold, we intentionally access this
        # attribute safely albeit somewhat more slowly via getattr().
        return getattr(hintable, '__annotations__', None)

    # ....................{ SETTERS                        }....................
    def set_pep649_hintable_annotations(
        # Mandatory parameters.
        hintable: Pep649Hintable,
        annotations: Pep649HintableAnnotations,

        # Optional parameters.
        exception_cls: TypeException = BeartypeDecorHintPep649Exception,
        exception_prefix: str = '',
    ) -> None:
        assert isinstance(annotations, dict), (
            f'{repr(annotations)} not dictionary.')
        assert all(
            isinstance(annotations_key, str) for annotations_key in annotations
        ), f'{repr(annotations)} not dictionary mapping names to type hints.'

        # If this hintable is *NOT* actually a hintable, raise an exception.
        # Amusingly, the simplest means of implementing this validation is to
        # simply retrieve the prior "__annotations__" dunder dictionary
        # currently set on this hintable.
        get_pep649_hintable_annotations(
            hintable=hintable,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )
        # Else, this hintable is actually a hintable.

        # Attempt to...
        try:
            # Atomically (i.e., all-at-once) replace this hintable's existing
            # "__annotations__" dunder dictionary with these new annotations. Do
            # so atomically for both safety and efficiency.
            hintable.__annotations__ = annotations
        # If doing so fails with an exception resembling the following, this
        # hintable is *NOT* pure-Python. The canonical example are C-based
        # decorator objects (e.g., class, property, or static method
        # descriptors), whose exception message reads:
        #     AttributeError: 'method' object has no attribute '__annotations__'
        #
        # C-based decorator objects define a read-only "__annotations__" dunder
        # attribute that proxies an original writeable "__annotations__" dunder
        # attribute of the pure-Python callables they originally decorated.
        # Detecting this edge case is non-trivial and most easily deferred to
        # this late time. While non-ideal, simplicity >>>> idealism here.
        except AttributeError:
            # For the name of each annotated attribute of this hintable and the
            # new hint which which to annotate this attribute, overwrite the
            # prior hint originally annotating this attribute with this new
            # hint.
            #
            # Note that:
            # * The above assignment is an efficient O(1) operation and thus
            #   intentionally performed first.
            # * This iteration-based assignment is an inefficient O(n) operation
            #   (where "n" is the number of annotated attributes of this
            #   hintable) and thus intentionally performed last here.
            for attr_name, attr_hint in annotations.items():
                hintable.__annotations__[attr_name] = attr_hint

# ....................{ VERSIONS ~ docs                    }....................
get_pep649_hintable_annotations_or_none.__doc__ = (
    '''
    **Immutable memoized annotations** (i.e., :class:`.FrozenDict` instance
    encapsulating the possibly empty ``__annotations__`` dunder dictionary
    mapping from the name of each annotated child object of the passed hintable
    to the type hint annotating that child object) annotating the passed
    **hintable** (i.e., ideally pure-Python object defining the
    ``__annotations__`` dunder attribute as well as the :pep:`649`-compliant
    ``__annotate__`` dunder method if the active Python interpreter targets
    Python >= 3.14) if this hintable defines the ``__annotations__`` dunder
    dictionary *or* :data:`None` otherwise (i.e., if this hintable fails to
    define the ``__annotations__`` dunder dictionary).

    This getter is memoized for efficiency, guaranteeing amortized worst-case
    :math:`O(1)` constant time complexity. The first call to this getter passed
    a new hintable annotated by one or more type hints containing :math:`n`
    unquoted forward references exhibits non-amortized worst-case :math:`O(n)`
    linear time complexity, justifying the memoization of this getter.

    This getter is memoized *only* under Python >= 3.14. Why? Because the
    lower-level :func:`annotationlib.get_annotations` getter underlying this
    higher-level getter *only* memoizes the annotations dictionary it creates
    and returns when passed the ``format=Format.VALUE`` keyword parameter. When
    passed *any* other ``format`` value, :func:`annotationlib.get_annotations`
    avoids avoids caching its return value. Creating this return value is
    algorithmically non-trivial and expensive. Sadly, we are effectively
    required to memoize this return value here.

    Parameters
    ----------
    hintable : Pep649Hintable
        Hintable to be inspected.
    hint_format : Format, default: Format.FORWARDREF
        Format of annotated hints to be returned. Defaults to
        :attr:`Format.FORWARDREF`, in which case this getter safely encapsulates
        each otherwise unsafe unquoted forward reference transitively
        subscripting each hint annotating this hintable with a safe
        :class:`annotationlib.ForwardRef` object. See also the higher-level
        :func`.get_pep649_hintable_annotations` getter for further details.
    exception_cls : TypeException, default: BeartypeDecorHintPep649Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep649Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    Optional[Pep649HintableAnnotations]
        Either:

        * If this hintable is actually a hintable, the ``__annotations__``
          dunder dictionary set on this hintable.
        * Else, :data:`None`.
    '''
)
set_pep649_hintable_annotations.__doc__ = (
    '''
    Set the **annotations** (i.e., ``__annotations__`` dunder dictionary mapping
    from the name of each annotated child object of the passed hintable to the
    type hint annotating that child object) annotating the passed **hintable**
    (i.e., ideally pure-Python object defining the ``__annotations__`` dunder
    attribute as well as the :pep:`649`-compliant ``__annotate__`` dunder method
    if the active Python interpreter targets Python >= 3.14) to the passed
    dictionary.

    Caveats
    -------
    **This setter preserves unmodified the existing** :attr:`Format.VALUE`
    **and** :attr:`Format.STRING` **formats of the** ``__annotations__``
    **dunder dictionary of the passed hintable,** as originally created and
    returned by the original ``__annotate__`` dunder method bound to this
    hintable. This setter *only* modifies the :attr:`Format.FORWARDREF` format.
    Why? Because there exist two distinct cases, which although distinct imply
    the same conclusion:

    * When the caller of an ``__annotate__`` dunder method passes the
      :attr:`Format.VALUE` format, they expect that method to raise a
      :exc:`NameError` exception if the ``__annotations__`` dunder dictionary
      underlying that call contains one or more unquoted forward references.
      Indeed, this is the *only* efficient (and thus reasonable) means of
      detecting whether a hintable is annotated by unquoted forward references.
      This is also the only valid use case for passing the :attr:`Format.VALUE`
      format. Although this valid use case is of marginal utility, it is still
      of utility and *must* be preserved as such. But the passed ``annotations``
      dictionary exists (rather than raising a :exc:`NameError` exception) and
      thus contains *no* unquoted forward references! Ergo, this ``annotations``
      dictionary *cannot* be returned if the caller passes the
      :attr:`Format.VALUE` format. Doing so would destroy this format's only
      valid use case, which can only be preserved by deferring to the original
      ``__annotate__`` dunder method bound to this hintable.
    * When the caller of an ``__annotate__`` dunder method passes the
      :attr:`Format.STRING` format, they expect that method to return
      human- and machine-readable string representations of the *original* type
      hints annotating this hintable. These strings are expected to be readably
      concise and machine-comparable. These strings are, in particular, *not*
      expected to contain **forward reference proxies** (e.g., either standard
      :class:`annotationlib.ForwardRef` objects or non-standard
      beartype-specific objects behaving similarly). Forward reference proxies
      typically have verbose string representations, confounding
      human-readability. They also do *not* necessarily compare equal to other
      objects, confounding machine-comparability. In short, the string
      representation of ``__annotations__`` dunder dictionaries should contain
      *no* forward reference proxies. However, the dictionaries passed to this
      setter often contain forward reference proxies! Ergo, this ``annotations``
      dictionary *cannot* be returned if the caller passes the
      :attr:`Format.STRING` format. Doing so would destroy this format's most
      common use cases, which can only be preserved by deferring to the original
      ``__annotate__`` dunder method bound to this hintable.

      More generally (and ignoring the above concerns about forward reference
      proxies), we can say that callers requesting documentation are ultimately
      requesting human-readable string representations of the *original* type
      hints annotating this hintable. Those type hints are what the third-party
      packages defining those hintables intended those hintables to be annotated
      as. Those type hints embody those intentions, thus constituting the most
      readable description of those hintables.

    **This setter replaces the original** ``__annotate__`` **dunder method bound
    to this hintable,** monkey-patching that method with a new ``__annotate__``
    dunder method that returns either:

    * If the caller passed the :attr:`Format.FORWARDREF` format, the
      ``annotations`` parameter passed to this higher-level setter.
    * Else (e.g., if the caller passed either the :attr:`Format.VALUE` or
      :attr:`Format.STRING` formats), the result of calling the original
      ``__annotate__`` dunder method bound to this hintable.

    Parameters
    ----------
    hintable : Pep649Hintable
        Hintable to be inspected.
    annotations : Pep649HintableAnnotations
        ``__annotations__`` dunder dictionary to set on this hintable.
    exception_cls : TypeException, default: BeartypeDecorHintPep649Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep649Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
         If this hintable fails to define the ``__annotations__`` dunder
         dictionary. Since *all* pure-Python hintables (including unannotated
         hintables) define this dictionary, this getter raises an exception only
         if the passed hintable is either:

         * *Not* actually a hintable.
         * A **pseudo-callable object** (i.e., otherwise uncallable object whose
           class renders all instances of that class callable by defining the
           ``__call__()`` dunder method).
    '''
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Typed object pool** (i.e., submodule whose thread-safe API caches previously
instantiated objects of arbitrary types for space- and time-efficient reuse by
the :func:`beartype.beartype` decorator across decoration calls).

This private submodule is *not* intended for importation by downstream callers.

Caveats
-------
**This submodule only pools objects defining an** ``__init__()`` **method
accepting no parameters.** Why? Because this submodule unconditionally pools
all objects of the same types under those types. This submodule provides *no*
mechanism for pooling objects of the same types under different parameters
instantiated with those parameters and thus only implements a coarse- rather
than fine-grained object cache. If insufficient, consider defining a new
submodule implementing a fine-grained object cache unique to those objects. For
example:

* This submodule unconditionally pools all instances of the
  :class:`beartype._check.metadata.metadecor.BeartypeDecorMeta` class under that
  type.
* The parallel :mod:`beartype._util.cache.pool.utilcachepoollistfixed`
  submodule conditionally pools every instance of the
  :class:`beartype._util.cache.pool.utilcachepoollistfixed.FixedList` class of
  the same length under that length.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Any,
    Type,
)
from beartype._data.cls.datacls import TYPES_BUILTIN_CONTAINER_MUTABLE
from beartype._data.typing.datatyping import T
from beartype._util.cache.pool.utilcachepool import KeyPool

# ....................{ (ACQUIRERS|RELEASERS)              }....................
def acquire_instance(cls: Type[T]) -> T:
    '''
    Acquire an arbitrary object of the passed type.

    If this type is that of a **builtin mutable container** (i.e.,
    :class:`bytearray`, :class:`dict`, :class:`list`, :class:`set`), this
    function guarantees the returned container to be empty. Specifically, this
    function internally calls the ``clear()`` method of this container *before*
    returning this container.

    Caveats
    -------
    **The contents of this object are otherwise arbitrary.** Aside from the
    aforementioned call to the ``clear()`` method of builtin mutable containers,
    callers should make *no* assumptions as to this object's state. Instead,
    callers should reinitialize this object immediately after acquiring this
    object.

    Parameters
    ----------
    cls : Type[T]
        Type of the object to be acquired.

    Returns
    -------
    T
        Arbitrary object of this type.
    '''
    assert isinstance(cls, type), f'{repr(cls)} not type.'

    # Thread-safely acquire an object of this type.
    instance = _instance_pool_acquire(cls)  # type: ignore[arg-type]
    # assert isinstance(instance, cls), (
    #     '{!r} not a {!r}.'.format(instance, cls))

    # If this type is that of a builtin mutable container (i.e., "bytearray",
    # "dict", "list", "set"), safely empty this container on behalf of the
    # caller. Since all callers requesting a builtin mutable container prefer
    # this container to be empty *AND* since detecting builtin mutable container
    # types is trivially fast, we do so on behalf of the caller.
    if cls in TYPES_BUILTIN_CONTAINER_MUTABLE:
        instance.clear()  # type: ignore[attr-defined]
    # Else, this type is *NOT* that of a builtin mutable container. In this
    # case, preserve this instance as is.

    # Return this object.
    return instance  # type: ignore[return-value]


def release_instance(obj: Any) -> None:
    '''
    Release the passed object acquired by a prior call to the
    :func:`acquire_instance` function.

    Caveats
    -------
    **This object is not safely accessible after calling this function.**
    Callers should make *no* attempts to read, write, or otherwise access this
    object, but should instead nullify *all* variables referring to this object
    immediately after releasing this object (e.g., by setting these variables
    to the ``None`` singleton *or* by deleting these variables).

    Parameters
    ----------
    obj : object
        Previously acquired object to be released.
    '''

    # Thread-safely release this object.
    #
    # Note that we intentionally pass parameters positionally rather than by
    # keyword as a negligible microoptimization.
    _instance_pool_release(obj, obj.__class__)

# ....................{ SINGLETONS ~ private               }....................
_instance_pool = KeyPool(item_maker=lambda cls: cls())
'''
Thread-safe **typed object pool** (i.e., :class:`.KeyPool` singleton caching
previously instantiated objects of the same types under those types).

Caveats
-------
**Avoid accessing this private singleton externally.** Instead, call the public
:func:`acquire_instance` and :func:`release_instance` functions, which
efficiently validate both input *and* output to conform to sane expectations.
'''


_instance_pool_acquire = _instance_pool.acquire
'''
:meth:`.KeyPool.acquire` method bound to the :data:`._object_type_pool`,
globalized as a negligible microoptimization.
'''


_instance_pool_release = _instance_pool.release
'''
:meth:`.KeyPool.release` method bound to the :data:`._object_type_pool`,
globalized as a negligible microoptimization.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **abstract syntax tree (AST) tester utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.ast.utilasttest` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_is_node_callable_typed() -> None:
    '''
    Test the :func:`beartype._util.ast.utilasttest.is_node_callable_typed`
    factory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.ast.utilastmake import make_node_from_code_snippet
    from beartype._util.ast.utilasttest import is_node_callable_typed

    # ....................{ LOCALS                         }....................
    # Tuple of 2-tuples "(is_node_callable_typed_bool, func_code)" describing
    # example callables to be tested below, where:
    # * "is_node_callable_typed_bool" is the boolean value expected to be
    #   returned by the is_node_callable_typed() tester when passed the
    #   "ast.Callable" AST tree parsed from the "func_code" item.
    # * "func_code" is a triple-quoted string defining an arbitrary callable to
    #   be parsed into the "ast.Callable" AST tree to be passed to the
    #   is_node_callable_typed() tester.
    CALLABLE_DATAS = (
        # Untyped function accepting arbitrary parameters.
        (False, '''def _(a, b, /, c, d, *args, e, f, **kwargs): ...''',),
        # Typed function with an annotated return and unannotated parameters.
        (True, '''def _(a, b, /, c, d, *args, e, f, **kwargs) -> None: ...''',),
        # Typed function with an annotated non-variadic flexible parameter,
        # other unannotated parameters, and an unannotated return.
        (True, '''def _(a, b: int, /, c, d, *args, e, f, **kwargs): ...''',),
        # Typed function with an annotated non-variadic keyword-only parameter
        # defined implicitly, other unannotated parameters, and an unannotated
        # return.
        (True, '''def _(a, b, /, c, d, *args, e, f: bool, **kwargs): ...''',),
        # Typed function with an annotated non-variadic keyword-only parameter
        # defined explicitly, other unannotated parameters, and an unannotated
        # return.
        (True, '''def _(a, b, /, c, d, *, e, f: bool, **kwargs): ...''',),
        # Typed function with an annotated non-variadic positional-only
        # parameter defined explicitly, other unannotated parameters, and an
        # unannotated return.
        (True, '''def _(a, b, /, c, d: float, *, d, e, **kwargs): ...''',),
        # Typed function with an annotated positional variadic argument
        # parameter, other unannotated parameters, and an unannotated return.
        (True, '''def _(a, b, /, c, d, *args: str, e, f, **kwargs): ...''',),
        # Typed function with an annotated keyword variadic argument
        # parameter, other unannotated parameters, and an unannotated return.
        (True, '''def _(a, b, /, c, d, *args, e, f, **kwargs: bytes): ...''',),
    )

    # ....................{ ASSERTS                        }....................
    # For each example callable to be tested...
    for is_node_callable_typed_bool, func_code in CALLABLE_DATAS:
        # "ast.Callable" node parsed from the string defining this callable.
        func_node = make_node_from_code_snippet(func_code)

        # Assert that this tester reports the expected boolean when passed this
        # "ast.Callable" node.
        assert is_node_callable_typed(func_node) is is_node_callable_typed_bool



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype exception data** (i.e., high-level globals and constants leveraged
throughout the :mod:`beartype._check.error` subpackage).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Callable,
    Dict,
    Optional,
)
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._check.error.errcause import ViolationCause

# ....................{ GLOBALS                            }....................
# Initialized with automated inspection below in the _init() function.
HINT_SIGN_TO_GET_CAUSE_FUNC: Dict[
    Optional[HintSign], Callable[[ViolationCause], ViolationCause]] = {}
'''
Dictionary mapping each **sign** (i.e., arbitrary object uniquely identifying a
category of type hints) to a private getter function defined by this submodule
whose signature matches that of the :func:`._find_cause` function and
which is dynamically dispatched by that function to describe type-checking
failures specific to that unsubscripted :mod:`typing` attribute.
'''

# ....................{ PRIVATE ~ initializers             }....................
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer heavyweight imports.
    from beartype._data.hint.sign.datahintsigns import (
        HintSignAnnotated,
        HintSignForwardRef,
        HintSignLiteral,
        HintSignNoReturn,
        HintSignPep484585GenericUnsubbed,
        HintSignPep484585TupleFixed,
        HintSignType,
    )
    from beartype._data.hint.sign.datahintsignset import (
        HINT_SIGNS_MAPPING,
        HINT_SIGNS_ORIGIN_ISINSTANCEABLE,
        HINT_SIGNS_CONTAINER_ARGS_1,
        HINT_SIGNS_UNION,
    )
    from beartype._check.error._nonpep.errnonpeptype import (
        find_cause_instance_type_forwardref,
        find_cause_nonpep,
        find_cause_type_instance_origin,
    )
    from beartype._check.error._pep.errpep484604 import (
        find_cause_pep484604_union)
    from beartype._check.error._pep.errpep586 import find_cause_pep586_literal
    from beartype._check.error._pep.errpep593 import find_cause_pep593_annotated
    from beartype._check.error._pep.pep484.errpep484noreturn import (
        find_cause_pep484_noreturn)
    from beartype._check.error._pep.pep484585.errpep484585container import (
        find_cause_pep484585_container_args_1,
        find_cause_pep484585_tuple_fixed,
    )
    from beartype._check.error._pep.pep484585.errpep484585generic import (
        find_cause_pep484585_generic_unsubbed)
    from beartype._check.error._pep.pep484585.errpep484585mapping import (
        find_cause_pep484585_mapping)
    from beartype._check.error._pep.pep484585.errpep484585subclass import (
        find_cause_pep484585_subclass)

    # ....................{ FALLBACKS                      }....................
    # Map each originative sign to the appropriate finder *BEFORE* any other
    # mappings. This is merely a generalized fallback subsequently replaced by
    # sign-specific finders below.
    for hint_sign in HINT_SIGNS_ORIGIN_ISINSTANCEABLE:
        HINT_SIGN_TO_GET_CAUSE_FUNC[hint_sign] = find_cause_type_instance_origin

    # Map each 1-argument container sign to its corresponding finder.
    for hint_sign in HINT_SIGNS_CONTAINER_ARGS_1:
        HINT_SIGN_TO_GET_CAUSE_FUNC[hint_sign] = (
            find_cause_pep484585_container_args_1)

    # Map each 2-argument mapping sign to its corresponding finder.
    for hint_sign in HINT_SIGNS_MAPPING:
        HINT_SIGN_TO_GET_CAUSE_FUNC[hint_sign] = find_cause_pep484585_mapping

    # Map each union-specific sign to its corresponding finder.
    for hint_sign in HINT_SIGNS_UNION:
        HINT_SIGN_TO_GET_CAUSE_FUNC[hint_sign] = find_cause_pep484604_union

    # ....................{ SPECIFICS                      }....................
    # Map each sign validated by a unique finder to that finder *AFTER* all
    # other mappings. These sign-specific finders are intended to replace all
    # other automated mappings above.
    HINT_SIGN_TO_GET_CAUSE_FUNC.update({
        # ....................{ NON-PEP                    }....................
        # Map PEP-noncompliant hints identified by *NO* PEP-compliant signs to
        # the catch-all PEP-noncompliant cause finder.
        None: find_cause_nonpep,

        # ....................{ PEP 484                    }....................
        HintSignForwardRef: find_cause_instance_type_forwardref,
        HintSignNoReturn: find_cause_pep484_noreturn,

        # ....................{ PEP (484|585)              }....................
        HintSignPep484585GenericUnsubbed: (
            find_cause_pep484585_generic_unsubbed),
        HintSignPep484585TupleFixed: find_cause_pep484585_tuple_fixed,
        HintSignType: find_cause_pep484585_subclass,

        # ....................{ PEP 586                    }....................
        HintSignLiteral: find_cause_pep586_literal,

        # ....................{ PEP 593                    }....................
        HintSignAnnotated: find_cause_pep593_annotated,
    })


# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **beartype import hookable** :pep:`557` **submodule** (i.e., data
module containing *only* :pep:`557`-compliant dataclass declarations, mimicking
real-world usage of the :func:`beartype.claw.beartype_package` import hook on an
external package declaring these dataclasses).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintParamViolation
# from beartype.typing import (
#     List,
#     Union,
# )
from dataclasses import (
    dataclass,
    field,
)
from pytest import raises

# from beartype.claw._clawstate import claw_state
# print(f'this_submodule conf: {repr(claw_state.module_name_to_beartype_conf)}')

# ....................{ PEP 557                            }....................
@dataclass
class ToRenderUpThyCharge(object):
    '''
    Arbitrary dataclass implicitly decorated by the :func:`beartype.beartype`
    decorator with non-default parameters.
    '''

    # This annotated field declaration is safely type-checkable by
    # die_if_unbearable(), clearly.
    and_though_neer_yet: str = 'Thou hast unveiled thy inmost sanctuary,'

    # This annotated field declaration is *NOT* safely type-checkable by
    # die_if_unbearable(). Clearly, a dataclass "field" instance is *NOT* a
    # valid string and thus violates the type hint annotating this field. Since
    # PEP 681 standardizes declarations like this as semantically valid,
    # beartype has *NO* alternative but to quietly turn a blind eye to what
    # otherwise might be considered a type violation.
    and_twilight_phantasms: str = field(
        default='Enough from incommunicable dream,')


# Assert that instantiating this dataclass both with *AND* without passing valid
# initialization parameters succeeds as expected.
and_deep_noonday_thought = ToRenderUpThyCharge()
has_shone_within_me = ToRenderUpThyCharge(
    and_though_neer_yet='And moveless, as a long-forgotten lyre',
    and_twilight_phantasms='Suspended in the solitary dome',
)

# Assert that instantiating this dataclass with invalid initialization
# parameters raises the expected exception.
with raises(BeartypeCallHintParamViolation):
    ToRenderUpThyCharge(
        and_though_neer_yet=b'Of some mysterious and deserted fane,')
with raises(BeartypeCallHintParamViolation):
    ToRenderUpThyCharge(and_twilight_phantasms=(
        b'I wait thy breath, Great Parent, that my strain'))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator return code generator** (i.e., low-level callables
dynamically generating Python expressions type-checking the annotated return of
the callable currently being decorated by the :func:`beartype.beartype`
decorator in a general-purpose manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._check.checkmake import (
    make_code_raiser_func_pith_check,
    make_code_raiser_func_pep484_noreturn_check,
)
from beartype._check.convert.convmain import sanify_hint_root_func
from beartype._check.metadata.hint.hintsane import HINT_SANE_IGNORABLE
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._data.code.pep.datacodepep484 import PEP484_CODE_CHECK_NORETURN
from beartype._data.error.dataerrmagic import EXCEPTION_PLACEHOLDER
from beartype._data.func.datafuncarg import (
    ARG_NAME_RETURN,
    ARG_NAME_RETURN_REPR,
)
from beartype._data.typing.datatyping import LexicalScope
from beartype._data.typing.datatypingport import Hint
from beartype._data.code.datacodefunc import (
    CODE_RETURN_CHECK_PREFIX_format,
    CODE_RETURN_CHECK_SUFFIX_format,
)
from beartype._decor._nontype._wrap._wraputil import unmemoize_func_wrapper_code
from beartype._util.error.utilerrraise import reraise_exception_placeholder
from beartype._util.error.utilerrwarn import reissue_warnings_placeholder
from beartype._util.hint.utilhinttest import is_hint_needs_cls_stack
from beartype._util.kind.maplike.utilmapset import update_mapping
from beartype._util.text.utiltextprefix import prefix_callable_return
from beartype._data.kind.datakindiota import SENTINEL
from typing import NoReturn
from warnings import catch_warnings

# ....................{ CODERS                             }....................
def code_check_return(decor_meta: BeartypeDecorMeta) -> str:
    '''
    Generate a Python code snippet type-checking the annotated return declared
    by the decorated callable if any *or* the empty string otherwise (i.e., if
    this return is unannotated).

    Parameters
    ----------
    decor_meta : BeartypeDecorMeta
        Decorated callable to be type-checked.

    Returns
    -------
    str
        Code type-checking any annotated return of the decorated callable.

    Raises
    ------
    BeartypeDecorHintPep484585Exception
        If this callable is either:

        * A coroutine *not* annotated by a :obj:`typing.Coroutine` type hint.
        * A generator *not* annotated by a :obj:`typing.Generator` type hint.
        * An asynchronous generator *not* annotated by a
          :obj:`typing.AsyncGenerator` type hint.
    BeartypeDecorHintNonpepException
        If the type hint annotating this return (if any) of this callable is
        neither:

        * **PEP-compliant** (i.e., :mod:`beartype`-agnostic hint compliant with
          annotation-centric PEPs).
        * **PEP-noncompliant** (i.e., :mod:`beartype`-specific type hint *not*
          compliant with annotation-centric PEPs)).
    '''
    assert isinstance(decor_meta, BeartypeDecorMeta), (
        f'{repr(decor_meta)} not beartype call.')

    # ..................{ LOCALS                             }..................
    # Possibly insane hint annotating this callable's return if any *OR* the
    # sentinel placeholder otherwise (i.e., if this return is unannotated).
    #
    # Note that "None" is a semantically meaningful PEP 484-compliant hint
    # equivalent to "type(None)". Ergo, we *MUST* explicitly distinguish between
    # "None" and an unannotated return with a sentinel.
    hint_insane: Hint = decor_meta.func_annotations_get(  # type: ignore[assignment]
        ARG_NAME_RETURN, SENTINEL)
    # print(f'func {decor_meta} return hint_insane: {hint_insane}')

    # If this return is unannotated, silently reduce to a noop.
    if hint_insane is SENTINEL:
        return ''
    # Else, this return is annotated.

    # Python code snippet to be returned, defaulting to the empty string
    # implying this callable's return to either be unannotated *OR* annotated by
    # a safely ignorable type hint.
    func_wrapper_code = ''

    # Lexical scope (i.e., dictionary mapping from the relative unqualified name
    # to value of each locally or globally scoped attribute accessible to a
    # callable or class), initialized to "None" for safety.
    func_scope: LexicalScope = None  # type: ignore[assignment]

    # ..................{ GENERATE                           }..................
    # Attempt to...
    try:
        # With a context manager "catching" *ALL* non-fatal warnings emitted
        # during this logic for subsequent "playrback" below...
        with catch_warnings(record=True) as warnings_issued:
            # Sanified hint metadata sanified from this possibly insane return.
            # If this hint is unsupported by @beartype, raise an exception.
            #
            # Note that:
            # * This hint must sanitized *BEFORE* testing this hint. Why? The
            #   PEP 484-compliant "typing.NoReturn" return hint in conjunction
            #   with PEP 563. If this return hint is "typing.NoReturn" *AND*
            #   this submodule enables PEP 563 via "from __future__ import
            #   annotations, then this hint will be the useless string
            #   "NoReturn" rather than the useful hint "typing.NoReturn". In
            #   this case, this hist *MUST* be sanitized first; doing so
            #   destringifies this string into a usable hint, enabling this hint
            #   to then be detected below.
            # * For the exact same reason, this sanitization *CANNOT* be
            #   performed in the low-level make_check_expr() dynamically
            #   generating code type-checking this hint.
            # print(f'Sanifying {repr(decor_meta)} return hint {repr(hint_insane)}...')
            hint_sane = sanify_hint_root_func(
                decor_meta=decor_meta,
                hint=hint_insane,
                pith_name=ARG_NAME_RETURN,
                exception_prefix=EXCEPTION_PLACEHOLDER,
            )
            # print(f'Sanified {repr(decor_meta)} return hint {repr(hint_insane)} to {repr(hint_sane)}.')

            # If this is the PEP 484-compliant "typing.NoReturn" type hint
            # allowed *ONLY* as a return annotation...
            if hint_sane.hint is NoReturn:
                # Pre-generated code snippet validating this callable to *NEVER*
                # successfully return by unconditionally generating a violation.
                code_noreturn_check = PEP484_CODE_CHECK_NORETURN.format(
                    func_call_prefix=decor_meta.func_wrapper_code_call_prefix)

                # Code snippet handling the previously generated violation by
                # either raising that violation as a fatal exception or emitting
                # that violation as a non-fatal warning.
                (
                    code_noreturn_violation,
                    func_scope,
                    _
                ) = make_code_raiser_func_pep484_noreturn_check(decor_meta.conf)

                # Full code snippet to be returned.
                func_wrapper_code = (
                    f'{code_noreturn_check}{code_noreturn_violation}')
            # Else, this is *NOT* "typing.NoReturn".
            #
            # If this hint is unignorable...
            elif hint_sane is not HINT_SANE_IGNORABLE:
                #FIXME: DRY violation. The same logic appears in "_wrapargs" as
                #well. It looks like what we *PROBABLY* want to do here is:
                #* Rename the existing make_code_raiser_func_pith_check()
                #  factory to _make_code_raiser_func_pith_check_cached().
                #* Define a new make_code_raiser_func_pith_check() factory that
                #  is unmemoized and has a simpler public API. Notably, this new
                #  factory should:
                #  * Accept a new "decor_meta" parameter.
                #  * Drop the existing "conf" and "cls_stack" parameters.
                #  * Pass parameters by keyword rather than positionally. This
                #    would be especially useful for the "is_param" parameter,
                #    which would suddenly become readable below.
                #  * Internally compute the "cls_stack" as given below.

                # Type stack if required by this hint *OR* "None" otherwise.
                # See is_hint_needs_cls_stack() for details.
                #
                # Note that the original unsanitized "hint_insane" (e.g.,
                # "typing.Self") rather than the new sanitized "hint" (e.g.,
                # the class currently being decorated by @beartype) is
                # passed to that tester. See _code_check_args() for details.
                cls_stack = (
                    decor_meta.cls_stack
                    if is_hint_needs_cls_stack(hint_insane) else
                    None
                )
                # print(f'return hint {repr(hint_insane)} -> {repr(hint)} cls_stack: {repr(cls_stack)}')

                # Code snippet type-checking any arbitrary return.
                (
                    code_return_check_pith,
                    func_scope,
                    hint_refs_type_basename,
                ) = make_code_raiser_func_pith_check(  # type: ignore[assignment]
                    hint_sane,
                    decor_meta.conf,
                    cls_stack,
                    False,  # <-- True only for parameters
                )

                # Unmemoize this snippet against this return.
                code_return_check = unmemoize_func_wrapper_code(
                    decor_meta=decor_meta,
                    func_wrapper_code=code_return_check_pith,
                    pith_repr=ARG_NAME_RETURN_REPR,
                    hint_refs_type_basename=hint_refs_type_basename,
                )

                #FIXME: [SPEED] Optimize the following two string munging
                #operations into a single string-munging operation resembling:
                #    func_wrapper_code = CODE_RETURN_CHECK.format(
                #        func_call_prefix=decor_meta.func_wrapper_code_call_prefix,
                #        check_expr=code_return_check_pith_unmemoized,
                #    )
                #
                #Then define "CODE_RETURN_CHECK" in the "wrapsnip" submodule to
                #resemble:
                #    CODE_RETURN_CHECK = (
                #        f'{CODE_RETURN_CHECK_PREFIX}{{check_expr}}'
                #        f'{CODE_RETURN_CHECK_SUFFIX}'
                #    )

                # Code snippets prefixing and suffixing the type-checking of
                # this return.
                code_return_check_prefix = CODE_RETURN_CHECK_PREFIX_format(
                    func_call_prefix=decor_meta.func_wrapper_code_call_prefix)
                code_return_check_suffix = CODE_RETURN_CHECK_SUFFIX_format(
                    func_return_prefix=(
                        decor_meta.func_wrapper_code_return_prefix))

                # Full code snippet to be returned, consisting of:
                # * Calling the decorated callable and localize its return
                #   *AND*...
                # * Type-checking this return *AND*...
                # * Returning this return from this wrapper function.
                func_wrapper_code = (
                    f'{code_return_check_prefix}'
                    f'{code_return_check}'
                    f'{code_return_check_suffix}'
                )
            # Else, this hint is ignorable.
            # if not func_wrapper_code: print(f'Ignoring {decor_meta.func_name} return hint {repr(hint)}...')
        # If one or more warnings were issued, reissue these warnings with each
        # placeholder substring (i.e., "EXCEPTION_PLACEHOLDER" instance)
        # replaced by a human-readable description of this callable and
        # annotated return.
        if warnings_issued:
            reissue_warnings_placeholder(
                warnings=warnings_issued,
                target_str=prefix_callable_return(
                    func=decor_meta.func_wrappee,
                    is_color=decor_meta.conf.is_color,
                ),
            )
        # Else, *NO* warnings were issued.
    # If any exception was raised, reraise this exception with each placeholder
    # substring (i.e., "EXCEPTION_PLACEHOLDER" instance) replaced by a
    # human-readable description of this callable and annotated return.
    except Exception as exception:
        reraise_exception_placeholder(
            exception=exception,
            target_str=prefix_callable_return(
                func=decor_meta.func_wrappee,
                is_color=decor_meta.conf.is_color,
            ),
        )

    # ..................{ RETURN                             }..................
    # If a local scope is required to type-check this return, merge this scope
    # into the local scope currently required by the current wrapper function.
    if func_scope:
        update_mapping(decor_meta.func_wrapper_scope, func_scope)
    # Else, *NO* local scope is required to type-check this return.

    # Return this code.
    return func_wrapper_code



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484`-compliant **type variable violation describers** (i.e.,
functions returning human-readable strings explaining violations of
:pep:`484`-compliant :class:`typing.TypeVar` objects).

This private submodule is *not* intended for importation by downstream callers.
'''

#FIXME: Excise this submodule up, please. *sigh*
# # ....................{ IMPORTS                            }....................
# from beartype.typing import Optional
# from beartype._data.typing.datatypingport import Hint
# from beartype._data.hint.sign.datahintsigns import HintSignTypeVar
# from beartype._check.error.errcause import ViolationCause
# from beartype._util.hint.pep.proposal.pep484.pep484typevar import (
#     get_hint_pep484_typevar_bounded_constraints_or_none)
#
# # ....................{ GETTERS                            }....................
# def find_cause_pep484_typevar(cause: ViolationCause) -> ViolationCause:
#     '''
#     Output cause describing describing the failure of the decorated callable to
#     *not* return a value in violation of the passed **type variable** (i.e.,
#     :pep:`484`-compliant :class:`typing.TypeVar` object).
#
#     Parameters
#     ----------
#     cause : ViolationCause
#         Input cause providing this data.
#
#     Returns
#     -------
#     ViolationCause
#         Output cause type-checking this data.
#     '''
#     assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'
#     assert cause.hint_sign is HintSignTypeVar, (
#         f'{repr(cause.hint)} not "HintSignTypeVar".')
#
#     # ....................{ LOCALS                         }....................
#     # Hint mapped to by this type variable if one or more transitive parent
#     # hints previously mapped this type variable to a hint *OR* "None".
#     hint_child: Optional[Hint] = cause.typearg_to_hint.get(cause.hint)  # pyright: ignore
#
#     # ....................{ REDUCTION                      }....................
#     # If *NO* transitive parent hints previously mapped this type variable to a
#     # hint...
#     if hint_child is None:
#         # PEP-compliant hint synthesized from all bounded constraints
#         # parametrizing this type variable if any *OR* "None" otherwise.
#         #
#         # Note this call is intentionally passed positional rather positional
#         # keywords due to memoization.
#         hint_curr_bound = get_hint_pep484_typevar_bounded_constraints_or_none(
#             cause.hint, cause.exception_prefix)  # type: ignore[arg-type]
#
#         # If this type variable was parametrized by one or more bounded
#         # constraints, reduce this type variable to these bounded constraints.
#         if hint_curr_bound is not None:
#             hint_child = hint_curr_bound
#         # Else, this type variable was unparametrized. In this case, preserve
#         # this type variable as is.
#     # Else, one or more transitive parent hints previously mapped this type
#     # variable to a hint.
#
#     # If this type variable was either mapped to another hint by one or more
#     # transitive parent hints *OR* parametrized by one or more bounded
#     # constraints...
#     if hint_child is not None:
#         # Unignorable sane hint sanified from this possibly ignorable insane
#         # hint *OR* "None" otherwise (i.e., if this hint is ignorable).
#         hint_child = cause.sanify_hint_child(hint_child)
#     # Else, this type variable was neither mapped to another hint by one or more
#     # transitive parent hints *NOR* parametrized by one or more bounded
#     # constraints.
#
#     # If this type variable is reducible to an unignorable hint...
#     if hint_child is not None:
#         # Ignore this semantically useless type variable in favour of this
#         # semantically useful hint by replacing *ALL* hint metadata describing
#         # the former with the latter.
#         cause_return = cause.permute_cause(hint=hint_child).find_cause()
#     # Else, this type variable is *NOT* reducible to an unignorable hint. Since
#     # @beartype currently fails to generate type-checking code for type
#     # variables in and of themselves, type variables have *NO* intrinsic
#     # semantic meaning and are thus ignorable.
#     else:
#         cause_return = cause
#
#     # ....................{ RETURN                         }....................
#     # Return this output cause.
#     return cause_return



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Python version-agnostic signs** (i.e., instances of the
:class:`beartype._data.hint.sign.datahintsigncls.HintSign` class
uniquely identifying PEP-compliant type hints in a safe, non-deprecated manner
regardless of the Python version targeted by the active Python interpreter).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: Attributes imported here at module scope *MUST* be explicitly
# deleted from this module's namespace below.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.typing import (
    FrozenSet,
    List,
)
from beartype._data.hint.sign.datahintsigncls import HintSign as _HintSign

# ....................{ SIGNS ~ implicit : pep : (484|585) }....................
# User-defined generics, defined here rather than below to enable explicit
# "typing" exports signed below to trivially alias these generic signs.

HintSignPep484585GenericUnsubbed = _HintSign(
    name='Pep484585GenericUnsubscripted')
'''
Sign uniquely identifying all :pep:`484`- or :pep:`585`-compliant
**unsubscripted generics** (i.e., types subclassing either the
:pep:`484`-compliant :class:`typing.Generic` superclass, the
:pep:`544`-compliant :class:`typing.Protocol` superclass, or a
:pep:`585`-compliant type hint).
'''


HintSignPep484585GenericSubbed = _HintSign(
    name='Pep484585GenericSubscripted')
'''
Sign uniquely identifying all :pep:`484`- or :pep:`585`-compliant **subscripted
generics** (i.e., unsubscripted generic types originally parametrized by one or
more :pep:`484`-compliant type variables subscripted by a corresponding number
of arbitrary child type hints): e.g.,

.. code-block:: pycon

   >>> from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

   # Unsubscripted PEP 585 generic parametrized by a PEP 484 type variable.
   >>> class MuhGeneric[T](list[T]): pass
   >>> get_hint_pep_sign_or_none(MuhGeneric)
   HintSignPep484585GenericUnsubbed

   # Subscripted PEP 585 generic replacing that type variable with a type.
   >>> get_hint_pep_sign_or_none(MuhGeneric[int])
   HintSignPep484585GenericSubbed
'''

# ....................{ SIGNS ~ explicit : setup           }....................
_HINT_SIGNS_TYPING_LIST: List[_HintSign] = []
'''
List of the signs identifying *all* standard :mod:`typing` type hints and type
hint factories.

This private list only exists for a brief window of time. Specifically:

#. This list initializes the public :data:`HINT_SIGNS_TYPING` frozen set below.
#. This list is then immediately deleted to avoid polluting the module namespace
   with temporary globals.
'''


def _make_typing_hint_sign(name: str) -> _HintSign:
    '''
    Sign with an explicit analogue in the standard :mod:`typing` module.

    This factory additionally adds this sign to the :data:`.HINT_SIGNS_TYPING`
    frozen set of the signs identifying *all* standard :mod:`typing` type hints
    and type hint factories.

    Caveats
    -------
    **This higher-level factory should always be used in lieu of the
    lower-level** :class:`._HintSign` **class to instantiate signs identifying
    standard** :mod:`typing` **type hints and type hint factories.**

    Parameters
    ----------
    name : str
        Name of this :mod:`typing` sign.

    Returns
    -------
    _HintSign
        This :mod:`typing` sign.
    '''

    # Sign with this same.
    hint_sign = _HintSign(name)

    # Append this sign to this list.
    _HINT_SIGNS_TYPING_LIST.append(hint_sign)

    # Return this sign.
    return hint_sign

# ....................{ SIGNS ~ explicit : define          }....................
# Signs with explicit analogues in the standard "typing" module.
#
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: Signs defined by this module are synchronized with the "__all__"
# list global of the "typing" module bundled with the most recent CPython
# release. For that reason, these signs are:
# * Intentionally declared in the exact same order prefixed by the exact same
#   inline comments as for that list global.
# * Intentionally *NOT* commented with docstrings, both because:
#   * These docstrings would all trivially reduce to a single-line sentence
#     fragment resembling "Alias of typing attribute."
#   * These docstrings would inhibit diffing and synchronization by inspection.
# * Intentionally *NOT* conditionally isolated to the specific range of Python
#   versions whose "typing" module lists these attributes. For example, the
#   "HintSignAsyncContextManager" sign identifying the
#   "typing.AsyncContextManager" attribute that only exists under Python >=
#   3.7 could be conditionally isolated to that range of Python versions.
#   Technically, there exists *NO* impediment to doing so; pragmatically, doing
#   so would be ineffectual. Why? Because attributes *NOT* defined by the
#   "typing" module of the active Python interpreter cannot (by definition) be
#   used to annotate callables decorated by the @beartype decorator.
#
# When bumping beartype to support a new CPython release:
# * Declare one new attribute here for each new "typing" attribute added by
#   that CPython release regardless of whether beartype explicitly supports
#   that attribute yet. The subsequently called die_unless_hint_pep_supported()
#   validator will raise exceptions when passed these attributes.
# * Preserve attributes here that have since been removed from the "typing"
#   module in that CPython release to ensure their continued usability when
#   running beartype against older CPython releases.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# Super-special typing primitives.
HintSignAnnotated       = _make_typing_hint_sign('Annotated')
HintSignAny             = _make_typing_hint_sign('Any')
HintSignCallable        = _make_typing_hint_sign('Callable')
HintSignClassVar        = _make_typing_hint_sign('ClassVar')
HintSignConcatenate     = _make_typing_hint_sign('Concatenate')
HintSignFinal           = _make_typing_hint_sign('Final')
HintSignForwardRef      = _make_typing_hint_sign('ForwardRef')
# Generic  <-- ambiguous between subscripted and unsubscripted variants (disambiguated below)
HintSignLiteral         = _make_typing_hint_sign('Literal')

#FIXME: Excise this *AFTER* dropping Python <= 3.13 support. Python >= 3.14
#unifies all unions under "HintSignUnion", thankfully. Phew!
HintSignOptional        = _make_typing_hint_sign('Optional')
HintSignParamSpec       = _make_typing_hint_sign('ParamSpec')
HintSignParamSpecArgs   = _make_typing_hint_sign('ParamSpecArgs')
HintSignParamSpecKwargs = _make_typing_hint_sign('ParamSpecKwargs')
HintSignProtocol        = _make_typing_hint_sign('Protocol')

#FIXME: Rename all *UNAMBIGUOUS* references to "HintSignTuple" to
#"HintSignPep484585TupleVariadic" for disambiguity. Note, however, that some
#references to "HintSignTuple" are ambiguous (in the sense that it is unclear in
#that early context whether the tuple type hint in question refers to a fixed-
#or variadic-length tuple type hint). Ergo, this rename *CANNOT* be automated
#with a global regex but should be applied manually one-by-one. *sigh*

# Note that the name of this sign is intentionally the ambiguous name "Tuple"
# rather than the unambiguous name "Pep484585TupleVariadic". Why? Because...
# actually, we have no particularly good reason. For disambiguity, the name of
# this sign should ideally be the latter. Sadly, doing so currently induces
# spurious test failures. We can't bother to dissect this at the moment. Ergo,
# the unctuous status quo prevails. Laziness: "It's not always a virtue."
HintSignTuple = HintSignPep484585TupleVariadic = _make_typing_hint_sign('Tuple')
# HintSignTuple = HintSignPep484585TupleVariadic = _make_typing_hint_sign(
#     'Pep484585TupleVariadic')
'''
Sign uniquely identifying **variable-length tuple type hints,** including:

* :pep:`484`-compliant type hints of the form ``typing.Tuple[{hint_child_1},
  ...]`` where the last child type hint subscripting this parent hint is an
  ellipses (i.e., ``"..."`` string, :data:`Ellipses` singleton).
* :pep:`585`-compliant type hints of the form ``tuple[{hint_child_1}, ...]``
  where the last child type hint subscripting this parent hint is an ellipses
  (i.e., ``"..."`` string, :data:`Ellipses` singleton).

See Also
--------
HintSignPep484585TupleFixed
    Sign uniquely identifying **fixed-length tuple type hints.**
HintSignPep646TupleFixedVariadic
    Sign uniquely identifying mixed fixed-variadic tuple type hints.
'''

HintSignType         = _make_typing_hint_sign('Type')
HintSignTypeVar      = _make_typing_hint_sign('TypeVar')
HintSignTypeVarTuple = _make_typing_hint_sign('TypeVarTuple')
HintSignUnion        = _make_typing_hint_sign('Union')

# ABCs (from collections.abc).
HintSignAbstractSet         = _make_typing_hint_sign('AbstractSet')

#FIXME: Permanently remove this sign *AFTER* dropping support for Python 3.15.
HintSignByteString          = _make_typing_hint_sign('ByteString')

HintSignContainer           = _make_typing_hint_sign('Container')
HintSignContextManager      = _make_typing_hint_sign('ContextManager')
HintSignHashable            = _make_typing_hint_sign('Hashable')
HintSignItemsView           = _make_typing_hint_sign('ItemsView')
HintSignIterable            = _make_typing_hint_sign('Iterable')
HintSignIterator            = _make_typing_hint_sign('Iterator')
HintSignKeysView            = _make_typing_hint_sign('KeysView')
HintSignMapping             = _make_typing_hint_sign('Mapping')
HintSignMappingView         = _make_typing_hint_sign('MappingView')
HintSignMutableMapping      = _make_typing_hint_sign('MutableMapping')
HintSignMutableSequence     = _make_typing_hint_sign('MutableSequence')
HintSignMutableSet          = _make_typing_hint_sign('MutableSet')
HintSignSequence            = _make_typing_hint_sign('Sequence')
HintSignSized               = _make_typing_hint_sign('Sized')
HintSignValuesView          = _make_typing_hint_sign('ValuesView')
HintSignAwaitable           = _make_typing_hint_sign('Awaitable')
HintSignAsyncIterator       = _make_typing_hint_sign('AsyncIterator')
HintSignAsyncIterable       = _make_typing_hint_sign('AsyncIterable')
HintSignCoroutine           = _make_typing_hint_sign('Coroutine')
HintSignCollection          = _make_typing_hint_sign('Collection')
HintSignAsyncGenerator      = _make_typing_hint_sign('AsyncGenerator')
HintSignAsyncContextManager = _make_typing_hint_sign('AsyncContextManager')

# Structural checks, a.k.a. protocols.
HintSignReversible = _make_typing_hint_sign('Reversible')
# SupportsAbs   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsBytes   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsComplex   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsFloat   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsIndex   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsInt   <-- not a useful type hint (already an isinstanceable ABC)
# SupportsRound   <-- not a useful type hint (already an isinstanceable ABC)

# Concrete collection types.
HintSignChainMap = _make_typing_hint_sign('ChainMap')
HintSignCounter = _make_typing_hint_sign('Counter')
HintSignDeque = _make_typing_hint_sign('Deque')
HintSignDict = _make_typing_hint_sign('Dict')
HintSignDefaultDict = _make_typing_hint_sign('DefaultDict')
HintSignList = _make_typing_hint_sign('List')
HintSignOrderedDict = _make_typing_hint_sign('OrderedDict')
HintSignSet = _make_typing_hint_sign('Set')
HintSignFrozenSet = _make_typing_hint_sign('FrozenSet')
HintSignNamedTuple = _make_typing_hint_sign('NamedTuple')
HintSignTypedDict = _make_typing_hint_sign('TypedDict')
HintSignGenerator = _make_typing_hint_sign('Generator')

# Other concrete types.
HintSignMatch = _make_typing_hint_sign('Match')
HintSignPattern = _make_typing_hint_sign('Pattern')

# Other concrete type aliases.
# IO  <-- ambiguous between subscripted and unsubscripted variants
HintSignBinaryIO = HintSignPep484585GenericUnsubbed
HintSignTextIO = HintSignPep484585GenericUnsubbed

# One-off things.
# AnyStr   <-- not a unique type hint (merely a constrained "TypeVar")
# assert_never   <-- unusable as a type hint
# assert_type   <-- unusable as a type hint
# cast   <-- unusable as a type hint
# clear_overloads   <-- unusable as a type hint
# dataclass_transform   <-- unusable as a type hint
# final   <-- unusable as a type hint
# get_args   <-- unusable as a type hint
# get_origin   <-- unusable as a type hint
# get_type_hints   <-- unusable as a type hint
# is_protocol    <-- unusable as a type hint
# is_typeddict   <-- unusable as a type hint
HintSignLiteralString = _make_typing_hint_sign('LiteralString')
HintSignNever         = _make_typing_hint_sign('Never')
HintSignNewType       = _make_typing_hint_sign('NewType')
# no_type_check   <-- unusable as a type hint
# no_type_check_decorator   <-- unusable as a type hint
HintSignNoDefault     = _make_typing_hint_sign('NoDefault')

# Note that "NoReturn" is contextually valid *ONLY* as a top-level return hint.
# Since this use case is extremely limited, we explicitly generate code for this
# use case outside of the general-purpose code generation pathway for standard
# type hints. Since "NoReturn" is an unsubscriptable singleton, we explicitly
# detect this type hint with an identity test and thus require *NO* sign to
# uniquely identify this type hint.
#
# Theoretically, explicitly defining a sign uniquely identifying this type hint
# could erroneously encourage us to use that sign elsewhere; we should avoid
# that, as "NoReturn" is invalid in almost all possible contexts. Pragmatically,
# doing so nonetheless improves orthogonality when detecting and validating
# PEP-compliant type hints, which ultimately matters more than our subjective
# feelings about the matter. Wisely, we choose pragmatics.
#
# In short, "NoReturn" is insane.
HintSignNoReturn = _make_typing_hint_sign('NoReturn')

HintSignNotRequired     = _make_typing_hint_sign('NotRequired')
# overload   <-- unusable as a type hint
# override   <-- unusable as a type hint
HintSignParamSpecArgs   = _make_typing_hint_sign('ParamSpecArgs')
HintSignParamSpecKwargs = _make_typing_hint_sign('ParamSpecKwargs')
HintSignReadOnly        = _make_typing_hint_sign('ReadOnly')
HintSignRequired        = _make_typing_hint_sign('Required')
# reveal_type         <-- unusable as a type hint
# runtime_checkable   <-- unusable as a type hint
HintSignSelf            = _make_typing_hint_sign('Self')
# Text   <-- not actually a type hint (literal alias for "str")
# TYPE_CHECKING   <-- unusable as a type hint
HintSignTypeAlias       = _make_typing_hint_sign('TypeAlias')
HintSignTypeGuard       = _make_typing_hint_sign('TypeGuard')
HintSignTypeIs          = _make_typing_hint_sign('TypeIs')
# TypeAliasType  <-- not a unique type hint (merely a C-based type)
HintSignUnpack          = _make_typing_hint_sign('Unpack')

# Wrapper namespace for re type aliases.
#
# Note that "typing.__all__" intentionally omits the "Match" and "Pattern"
# attributes, which it oddly considers to comprise another namespace. *shrug*

# ....................{ SIGNS ~ explicit : teardown        }....................
HINT_SIGNS_TYPING: FrozenSet[_HintSign] = frozenset(_HINT_SIGNS_TYPING_LIST)
'''
Frozen set of all **typing signs** (i.e., identifying *all* standard
PEP-compliant :mod:`typing` type hints and type hint factories available in the
most recent stable CPython release).
'''

# ....................{ SIGNS ~ implicit                   }....................
# Signs with *NO* explicit analogues in the stdlib "typing" module but
# nonetheless standardized by one or more PEPs.

HintSignNone = _HintSign(name='None')
'''
Sign uniquely identifying the :data:`None` singleton, explicitly supported by
:pep:`484` but lacking an explicit analogue in the standard :mod:`typing`
module:

    When used in a type hint, the expression None is considered equivalent to
    type(None).
'''

# ....................{ SIGNS ~ implicit : lib             }....................
# Signs identifying PEP-noncompliant third-party type hints published by...
#
# ....................{ SIGNS ~ implicit : lib : numpy     }....................
HintSignNumpyArray = _HintSign(name='NumpyArray')   # <-- "numpy.typing.NDArray"
'''
...the :mod:`numpy.typing` subpackage.
'''

# ....................{ SIGNS ~ implicit : lib : pandera   }....................
HintSignPanderaAny = _HintSign(name='PanderaAny')   # <-- "pandera.typing.*"
'''
...the :mod:`pandera.typing` subpackage.

Specifically, define a single sign unconditionally matching *all* type hints
published by the :mod:`pandera.typing` subpackage. Why? Because Pandera insanely
publishes its own Pandera-specific PEP-noncompliant runtime type-checking
decorator :func:`pandera.check_types` that supports *only* Pandera-specific
PEP-noncompliant :mod:`pandera.typing` type hints. Since Pandera users are
already accustomed to decorating *all* Pandera-based callables (i.e., callables
accepting one or more parameters and/or returning one or more values which are
Pandera objects) by :func:`pandera.check_types`, attempting to type-check the
same objects already type-checked by that decorator would only inefficiently and
needlessly slow :mod:`beartype` down. Ergo, we ignore *all* Pandera type hints
by:

* Defining this catch-all singleton for Pandera type hints here.
* Denoting this singleton to be unconditionally ignorable elsewhere.
'''

# ....................{ SIGNS ~ implicit : pep : (484|585) }....................
HintSignPep484585TupleFixed = _HintSign(name='Pep484585TupleFixed')
'''
Sign uniquely identifying **fixed-length tuple type hints,** including:

* :pep:`484`-compliant type hints of the form ``typing.Tuple[{hint_child_1},
  ..., {hint_child_N}]`` where ``{hint_child_N}`` is *not* an ellipses (i.e.,
  ``"..."`` string, :data:`Ellipses` singleton).
* :pep:`585`-compliant type hints of the form ``tuple[{hint_child_1}, ...,
  {hint_child_N}]`` where ``{hint_child_N}`` is *not* an ellipses (i.e.,
  ``"..."`` string, :data:`Ellipses` singleton).

Note that:

* The ``"..."`` substring above is *not* a literal ellipses but simply denotes
  an arbitrary number of non-ellipses child type hints.
* The existing :data:`.HintSignTuple` sign uniquely identifies variable-length
  tuple type hints. Why? Because that sign naturally matches the unsubscripted
  :obj:`typing.Tuple` type hint factory, which is semantically equivalent to the
  ``typing.Tuple[object, ...]`` type hint, which is the widest possible
  variable-length tuple type hint.

See Also
--------
HintSignTuple
    Sign uniquely identifying variadic-length tuple type hints.
HintSignPep646TupleFixedVariadic
    Sign uniquely identifying mixed fixed-variadic tuple type hints.
'''

# ....................{ SIGNS ~ implicit : pep : 557       }....................
# dataclasses.InitVar[...].
HintSignPep557DataclassInitVar = _HintSign(name='Pep557DataclassInitVar')
'''
:pep:`557`-compliant :obj:`dataclasses.InitVar` type hint factory, annotating
class-scoped variable annotations of :func:`dataclass.dataclass`-decorated
data classes.
'''

# ....................{ SIGNS ~ implicit : pep : 585       }....................
# os.PathLike[...], weakref.weakref[...], et al.
HintSignPep585BuiltinSubscriptedUnknown = _HintSign(
    name='Pep585BuiltinSubscriptedUnknown')
'''
:pep:`585`-compliant C-based :class:`types.GenericAlias` superclass inheritable
by PEP-noncompliant pure-Python subclasses in either the standard library or
third-party packages, which when subscripted by otherwise PEP-compliant child
type hints produce PEP-noncompliant **unrecognized subscripted builtin type
hints** (i.e., C-based type hints that are *not* isinstanceable types,
instantiated by subscripting pure-Python origin classes unrecognized by
:mod:`beartype` and thus PEP-noncompliant).

Examples include:

* ``os.PathLike[...]`` type hints.
* ``weakref.weakref[...]`` type hints.

Unsurprisingly, :mod:`beartype` reduces C-based unrecognized subscripted builtin
type hints (which are *not* type-checkable as is) to their unsubscripted
pure-Python origin classes (which are type-checkable as is).
'''

# ....................{ SIGNS ~ implicit : pep : 646       }....................
#FIXME: Excise after generalizing our dynamic code generator for tuple hints to
#flexibly support PEP 646-compliant tuple hints. *sigh*

HintSignPep646TupleFixedVariadic = _HintSign(
    name='Pep646TupleFixedVariadic')
'''
Sign uniquely identifying :pep:`646`-compliant **mixed fixed-variadic tuple type
hints,** defined as hints of the form "tuple[{hint_child_1}, ...,
{hint_child_N}]" where:

* Exactly one "{hint_child_I}" for some :math:`1 <= I <= N` is either:

  * A :pep:`646`-compliant **unpacked type variable tuple** (i.e., type hint of
    either the implicit form "*T" *or* explicit form "typing.Unpack[T]" for an
    arbitrary Python identifier "T") .
  * A :pep:`646`-compliant **unpacked child tuple hint** (i.e., type hint of the
    form "*tuple[{hint_child_child_1}, ..., {hint_child_child_M}]").

* "{hint_child_N}" is *not* an ellipses (i.e., "..." string, :data:`Ellipses`
  singleton).

Note that the "..." substring above is *not* a literal ellipses but simply
denotes an arbitrary number of non-ellipses child type hints.

See Also
--------
HintSignTuple
    Sign uniquely identifying variadic-length tuple type hints.
HintSignPep484585TupleFixed
    Sign uniquely identifying fixed-length tuple type hints.
'''


HintSignPep646TupleUnpacked = _HintSign(name='Pep646TupleUnpacked')
'''
Sign uniquely identifying :pep:`646`-compliant **unpacked tuple type hints,**
defined as child tuple hints of the form "*tuple[{hint_child_child_1}, ...,
{hint_child_child_M}]" subscripting parent tuple hints of the form
"tuple[{hint_child_1}, ..., *tuple[{hint_child_child_1}, ...,
{hint_child_child_M}], ..., {hint_child_N}]".

Note that the ``"..."`` substring above is *not* a literal ellipses but simply
denotes an arbitrary number of non-ellipses child type hints.
'''


HintSignPep646TypeVarTupleUnpacked = _HintSign(
    name='Pep646TypeVarTupleUnpacked')
'''
Sign uniquely identifying :pep:`646`-compliant **unpacked type variable
tuples,** defined as child tuple hints of the form "*{typevartuple}" (where
"{typevartuple}" is an instance of the :class:`typing.TypeVarTuple` type)
subscripting parent tuple hints of the form
"tuple[{hint_child_1}, ..., *{typevartuple}, ..., {hint_child_N}]".

Note that the ``"..."`` substring above is *not* a literal ellipses but simply
denotes an arbitrary number of non-ellipses child type hints.
'''

# ....................{ SIGNS ~ implicit : pep : 692       }....................
HintSignPep692TypedDictUnpacked = _HintSign(name='Pep692TypedDictUnpacked')
'''
Sign uniquely identifying :pep:`692`-compliant **unpacked typed dictionaries,**
defined as type hints of the form "typing.Unpack[{typeddict}]" where
"{typeddict}" is a :class:`typing.TypedDict` subclass.
'''

# ....................{ SIGNS ~ implicit : pep : 695       }....................
# "type {alias_name}[{typevar_name}] = {alias_value}" statements.

HintSignPep695TypeAliasUnsubscripted = _HintSign(
    name='Pep695TypeAliasUnsubscripted')
'''
:pep:`695`-compliant C-based :class:`types.TypeAliasType` class of all
:pep:`695`-compliant **unsubscripted type aliases** (i.e., objects created as
the left-hand sides of statements of the form ``type {alias_name} =
{alias_value}``).

Most real-world type aliases are unsubscripted and thus identified by this sign.
'''


HintSignPep695TypeAliasSubscripted = _HintSign(
    name='Pep695TypeAliasSubscripted')
'''
Sign uniquely identifying all :pep:`695`-compliant **subscripted type aliases**
(i.e., unsubscripted type aliases originally parametrized by one or more
:pep:`484`-compliant type variables subscripted by a corresponding number of
arbitrary child type hints): e.g.,

.. code-block:: pycon

   >>> from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

   # Unsubscripted PEP 695 type alias parametrized by a PEP 484 type variable.
   >>> MuhTypeAlias[T] = T | float
   >>> get_hint_pep_sign_or_none(MuhTypeAlias)
   HintSignPep695TypeAliasUnsubscripted

   # Subscripted PEP 695 type alias replacing that type variable with a type.
   >>> get_hint_pep_sign_or_none(MuhTypeAlias[int])
   HintSignPep695TypeAliasSubscripted
'''

# ....................{ CLEANUP                            }....................
# Prevent all attributes imported above from polluting this namespace. Why?
# Logic elsewhere subsequently assumes a one-to-one mapping between the
# attributes of this namespace and signs.
del (
    FrozenSet,
    List,
    _HINT_SIGNS_TYPING_LIST,
    _HintSign,
    _make_typing_hint_sign,
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator type hint code generation unit tests.**

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to type-checking code dynamically generated by the
:mod:`beartype._decor._nontype._wrap.wrapmain` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ decorator                  }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Urgh!
# @ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
@ignore_warnings(DeprecationWarning)
def test_beartype(iter_hints_piths_meta) -> None:
    '''
    Test the :func:`beartype.beartype` decorator with respect to type-checking
    code dynamically generated by the
    :mod:`beartype._decor._nontype._wrap.wrapmain` submodule.

    This unit test effectively acts as a functional test and is thus the core
    test exercising decorator functionality from the end user perspective --
    the only perspective that matters in the end. Unsurprisingly, this test is
    mildly more involved than most. *Whatevah.*

    This test additionally attempts to avoid similar issues to a `prior issue
    <issue #5_>`__ of this decorator induced by repeated
    :func:`beartype.beartype` decorations of different callables annotated by
    one or more of the same PEP-compliant type hints.

    .. _issue #5:
       https://github.com/beartype/beartype/issues/5

    Parameters
    ----------
    iter_hints_piths_meta : Callable[[], Iterable[beartype_test.a00_unit.data.hint.util.data_hintmetautil.HintPithMetadata]]
        Factory function creating and returning a generator iteratively yielding
        ``HintPithMetadata`` instances, each describing a sample type hint
        exercising an edge case in the :mod:`beartype` codebase paired with a
        related object either satisfying or violating that hint.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintViolation
    from beartype._util.text.utiltextansi import strip_str_ansi
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        # HintPepMetadata,
        HintPithUnsatisfiedMetadata,
    )
    from beartype_test._util.pytroar import raises_uncached
    # from pytest import warns
    # from pytest import deprecated_call
    from re import search
    # from warnings import simplefilter

    # ....................{ MAIN                           }....................
    # Force pytest to temporarily allow deprecation warnings to be caught by the
    # warns() context manager for the duration of this test. By default, pytest
    # simply "passes through" all deprecation warnings for subsequent reporting
    # if tests otherwise successfully pass. Deprecation warnings include:
    # * "DeprecationWarning".
    # * "FutureWarning".
    # * "PendingDeprecationWarning".
    # simplefilter('always')

    # For each predefined type hint and associated metadata...
    for hint_pith_meta in iter_hints_piths_meta():
        # ....................{ LOCALS                     }....................
        # Metadata describing this hint.
        hint_meta = hint_pith_meta.hint_meta

        # Type hint to be type-checked.
        hint = hint_meta.hint

        # Object to type-check against this hint.
        pith = hint_pith_meta.pith

        # Metadata describing this pith.
        pith_meta = hint_pith_meta.pith_meta
        # print(f'Type-checking PEP type hint {repr(hint_meta.hint)}...')

        # Beartype decorator configured specifically for this hint.
        beartype_confed = beartype(conf=hint_meta.conf)

        def func_untyped(hint_param: hint) -> hint:
            '''
            Undecorated function both accepting a single parameter and returning
            a value annotated by this hint whose implementation trivially
            reduces to the identity function.
            '''

            return hint_param

        #FIXME: For unknown and probably uninteresting reasons, the
        #pytest.warns() context manager appears to be broken on our local
        #machine. We have no recourse but to unconditionally ignore this warning
        #at the module level. So much rage!
        #FIXME: It's likely this has something to do with the fact that Python
        #filters deprecation warnings by default. This is almost certainly a
        #pytest issue. Since this has become fairly unctuous, we should probably
        #submit a pytest issue report.
        #FIXME: Actually, pytest now appears to have explicit support for
        #testing that a code block emits a deprecation warning:
        #    with pytest.deprecated_call():
        #        myfunction(17)
        #See also: https://docs.pytest.org/en/6.2.x/warnings.html#ensuring-code-triggers-a-deprecation-warning
        #FIXME: Fascinatingly, warns() still refuses to capture warnings.
        #Although we certainly could call deprecated_call(), doing so is
        #stymied by the fact that
        #"BeartypeDecorHintPep585DeprecationWarning" does *NOT* subclass the
        #standard "DeprecationWarning" class. *sigh*
        #FIXME: *NONSENSE*! It's 2024 and we *STILL* can't get this to work.
        #It's been bloody years. It just doesn't work and there's *NO*
        #reasonable means for us to debug why. We've tried literally everything:
        #pytest.deprecated_call(), pytest.warns(), warnings.simplefilter().
        #Nuthin'. Pytest simply refuses to catch deprecation warnings. Kinda fed
        #up with the whole thing, honestly.
        #FIXME: *FASCINATING*. The comparable test_decor_pep613() unit test
        #successfully tests deprecation warnings emitted by the @beartype
        #decorator with the trivial pytest.warns() + warnings.simplefilter()
        #combo. So... why does doing so here fail? Some combination of iteration
        #and closures would seem to be the likely culprit. Notably, "pylint"
        #complains this about the above definition of func_untyped():
        #    cell-var-from-loop: Cell variable "hint" defined in loop
        #
        #It seems likely a similar issue is obstructing testing here.
        #FIXME: ...huh. This is Bizarro World stuff. We tried extricating this
        #logic into the comparable test_door_is_bearable() test, which also
        #iterates but *WITHOUT* defining closures. That test still fails to
        #capture deprecation warnings. So... it's iteration? Really? *OH*. The
        #iter_hints_piths_meta() iterator is itself defined as a closure of a
        #session-scoped fixture. Welp, that's almost certainly it. Something
        #about that fundamentally obstructs capturing of deprecation warnings.
        #Our only recourse now is to try:
        #* Defining a new test_hint_warnings() unit test.
        #* Manually iterating over type hints in that test *WITHOUT* calling
        #  iter_hints_piths_meta(). Instead, extricate the import aspects of
        #  iter_hints_piths_meta() into a simple "for" loop. Avoid anything
        #  extraneous (e.g., repetition, context managers).
        #FIXME: *THAT WAS IT*. Our new test_hint_warnings() unit test
        #successfully captures deprecation warnings. So. It's our session-scoped
        #iter_hints_piths_meta() fixture closure, huh? Well, at least that sorta
        #makes sense. This is *DEFINITELY* a pytest issue -- but good luck
        #reducing that one to a minimal-length reproducible working example.

        # # @beartype-generated wrapper function type-checking that function.
        # func_typed = None
        #
        # # If this beartype decorator emits a warning for this hint...
        # if hint_meta.warning_type is not None:
        #     with deprecated_call():
        #     # Decorate that function under a context manager asserting this
        #     # decoration to emit the expected warning.
        #     # with warns(hint_meta.warning_type):
        #         func_typed = beartype_confed(func_untyped)
        # # Else, this beartype decorator emits *NO* warning for this hint. In
        # # this case, decorate this callable as is.
        # else:
        #     func_typed = beartype_confed(func_untyped)
        func_typed = beartype_confed(func_untyped)

        # ....................{ VIOLATE                    }....................
        # If this pith violates this hint...
        if isinstance(pith_meta, HintPithUnsatisfiedMetadata):
            # ....................{ EXCEPTION ~ type       }....................
            # Assert this wrapper function raises the expected exception when
            # type-checking this pith against this hint.
            with raises_uncached(BeartypeCallHintViolation) as exception_info:
                func_typed(pith)

            # Exception captured by the prior call to this wrapper function.
            exception = exception_info.value

            # Exception type localized for debuggability. Sadly, the
            # pytest.ExceptionInfo.__repr__() dunder method fails to usefully
            # describe its exception metadata.
            exception_type = exception_info.type

            # Assert this exception metadata describes the expected exception
            # as a sanity check against upstream pytest issues and/or issues
            # with our raises_uncached() context manager.
            assert issubclass(exception_type, BeartypeCallHintViolation)

            # Assert this exception to be public rather than private. The
            # @beartype decorator should *NEVER* raise a private exception.
            assert exception_type.__name__[0] != '_'

            # ....................{ EXCEPTION ~ culprits   }....................
            # Tuple of the culprits responsible for this exception, localized to
            # avoid inefficiently recomputing these culprits on each access.
            exception_culprits = exception.culprits

            # Assert that this tuple both exists *AND* is non-empty.
            assert isinstance(exception_culprits, tuple)
            assert exception_culprits

            # First culprit, which *SHOULD* be either:
            # * If this pith can be weakly referenced, this pith.
            # * Else, this pith *CANNOT* be weakly referenced. In this case,
            #   fallback to the truncated representation of this pith.
            culprit_pith = exception_culprits[0]

            # If the first culprit is a string...
            if isinstance(culprit_pith, str):
                # Assert that this string is non-empty.
                assert culprit_pith
            # Else, the first culprit is *NOT* a string. In this case...
            else:
                # Assert that the first culprit is this pith.
                assert culprit_pith is pith

                # If two or more culprits were responsible...
                if len(exception_culprits) >= 2:
                    # For each culprit following the first...
                    for culprit_nonpith in exception_culprits[1:]:
                        # Assert that this subsequent culprit is *NOT* the pith.
                        assert culprit_nonpith is not pith

            # ....................{ EXCEPTION ~ message    }....................
            # Exception message raised by this wrapper function.
            exception_str = str(exception)
            exception_str = strip_str_ansi(exception_str)
            # print('exception message: {}'.format(exception_str))

            # Assert this exception message is prefixed by a substring
            # identifying the decorated callable with a human-readable label.
            assert exception_str.startswith('Function ')

            # Assert that iterables of uncompiled regular expression expected
            # to match and *NOT* match this message are *NOT* strings, as
            # commonly occurs when accidentally omitting a trailing comma in
            # tuples containing only one string: e.g.,
            # * "('This is a tuple, yo.',)" is a 1-tuple containing one string.
            # * "('This is a string, bro.')" is a string *NOT* contained in a
            #   1-tuple.
            assert not isinstance(
                pith_meta.exception_str_match_regexes, str)
            assert not isinstance(
                pith_meta.exception_str_not_match_regexes, str)

            # For each uncompiled regular expression expected to match this
            # message, assert this expression actually does so.
            #
            # Note that the re.search() rather than re.match() function is
            # called. The latter is rooted at the start of the string and thus
            # *ONLY* matches prefixes, while the former is *NOT* rooted at any
            # string position and thus matches arbitrary substrings as desired.
            for exception_str_match_regex in (
                pith_meta.exception_str_match_regexes):
                assert search(
                    exception_str_match_regex,
                    exception_str,
                ) is not None

            # For each uncompiled regular expression expected to *NOT* match
            # this message, assert this expression actually does so.
            for exception_str_not_match_regex in (
                pith_meta.exception_str_not_match_regexes):
                assert search(
                    exception_str_not_match_regex,
                    exception_str,
                ) is None
        # ....................{ SATISFY                    }....................
        # Else, this pith satisfies this hint. In this case...
        else:
            # Assert this wrapper function successfully type-checks this pith
            # against this hint *WITHOUT* modifying this pith.
            assert func_typed(pith) is pith



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable getters** (i.e., utility functions dynamically
querying and retrieving various properties of passed callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import (
    Callable,
    Optional,
)
from beartype._cave._cavefast import MethodBoundInstanceOrClassType
from beartype._data.typing.datatyping import (
    Pep649HintableAnnotations,
    TypeException,
)

# ....................{ GETTERS ~ descriptors              }....................
#FIXME: Unit test us up, please.
#FIXME: Docstring us up, please.
def get_func_boundmethod_self(
    # Mandatory parameters.
    func: MethodBoundInstanceOrClassType,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> object:
    '''
    Instance object to which the passed **C-based bound instance method
    descriptor** (i.e., callable implicitly instantiated and assigned on the
    instantiation of an object whose class declares an instance function (whose
    first parameter is typically named ``self``) as an instance variable of that
    object such that that callable unconditionally passes that object as the
    value of that first parameter on all calls to that callable) was bound as an
    instance attribute at instantiation time.

    Parameters
    ----------
    func : object
        Bound method descriptor to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilCallableWrapperException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    -------
    object
        Instance object to which this bound method descriptor is bound to.

    Raises
    ------
    exception_cls
         If the passed object is *not* a bound method descriptor.

    See Also
    --------
    :func:`beartype._util.func.utilfunctest.is_func_boundmethod`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import die_unless_func_boundmethod

    # If this object is *NOT* a class method descriptor, raise an exception.
    die_unless_func_boundmethod(
        func=func,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this object is a class method descriptor.

    # Return the pure-Python function wrapped by this descriptor. Just do it!
    return func.__self__



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`484` and :pep:`604` **type-checking expression snippets** (i.e.,
triple-quoted pure-Python string constants formatted and concatenated together
to dynamically generate boolean expressions type-checking arbitrary objects
against :pep:`484`- and :pep:`604`-compliant type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import CallableStrFormat

# ....................{ CODE                               }....................
CODE_PEP484604_UNION_PREFIX = '''('''
'''
:pep:`484`-compliant code snippet prefixing all code type-checking the current
pith against each subscripted argument of a :class:`typing.Union` type hint.
'''


CODE_PEP484604_UNION_SUFFIX = '''
{indent_curr})'''
'''
:pep:`484`-compliant code snippet suffixing all code type-checking the current
pith against each subscripted argument of a :class:`typing.Union` type hint.
'''


CODE_PEP484604_UNION_CHILD_NONPEP = '''
{{indent_curr}}    # True only if this pith is of one of these types.
{{indent_curr}}    isinstance({pith_curr_expr}, {hint_curr_expr}) or'''
'''
:pep:`484`-compliant code snippet type-checking the current pith against the
current PEP-noncompliant child argument subscripting a parent
:class:`typing.Union` type hint.

See Also
--------
:data:`CODE_PEP484604_UNION_CHILD_PEP`
    Further details.
'''


CODE_PEP484604_UNION_CHILD_PEP = '''
{{indent_curr}}    {hint_child_placeholder} or'''
'''
:pep:`484`-compliant code snippet type-checking the current pith against the
current PEP-compliant child argument subscripting a parent :class:`typing.Union`
type hint.

Caveats
-------
The caller is required to manually slice the trailing suffix ``" or"`` after
applying this snippet to the last subscripted argument of such a hint. While
there exist alternate and more readable means of accomplishing this, this
approach is the optimally efficient.

The ``{indent_curr}`` format variable is intentionally brace-protected to
efficiently defer its interpolation until the complete PEP-compliant code
snippet type-checking the current pith against *all* subscripted arguments of
this parent hint has been generated.
'''

# ....................{ FORMATTERS                         }....................
# str.format() methods, globalized to avoid inefficient dot lookups elsewhere.
# This is an absurd micro-optimization. *fight me, github developer community*
CODE_PEP484604_UNION_CHILD_PEP_format: CallableStrFormat = (
    CODE_PEP484604_UNION_CHILD_PEP.format)
CODE_PEP484604_UNION_CHILD_NONPEP_format: CallableStrFormat = (
    CODE_PEP484604_UNION_CHILD_NONPEP.format)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Sphinx** integration tests.

This submodule functionally tests the :func:`beartype.beartype` decorator with
respect to the third-party Sphinx documentation build toolchain.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import (
    skip,
    skip_unless_package,
)

# ....................{ TESTS                              }....................
#FIXME: *THIS IS HORRIFYING.* Clearly, requiring an obsolete Sphinx is becoming
#a pressing issue. With all due haste, un-skip this test *AFTER* updating our
#"pyproject.toml" to allow sane Sphinx versions. *gulp*
@skip('Currently broken due to @beartype requiring an obsolete Sphinx.')
@skip_unless_package('sphinx')
def test_sphinx_docs_other(tmp_path) -> None:
    '''
    Functional test validating that the :func:`beartype.beartype` decorator
    conditionally reduces to a noop when the active Python interpreter is
    building documentation for the third-party :mod:`sphinx` package.

    To do so, this test externally runs the ``sphinx-build`` command against a
    minimal-length Sphinx document tree exercising all known edge cases.

    Parameters
    ----------
    tmp_path : pathlib.Path
        Abstract path encapsulating a temporary directory unique to this unit
        test, created in the base temporary directory.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    # from beartype._util.api.external.utilsphinx import (
    #     _SPHINX_AUTODOC_SUBPACKAGE_NAME)
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)
    from beartype_test._util.command.pytcmdrun import run_command_forward_output
    from beartype_test._util.path.pytpathtest import (
        get_test_func_data_lib_sphinx_dir)

    # ..................{ SPHINX-BUILD                       }..................
    # Tuple of all shell words with which to run the "sphinx-build" command.
    SPHINX_ARGS = get_interpreter_command_words() + (
        # Fully-qualified name of the Sphinx package to be run. Interestingly,
        # running this package as a script is semantically equivalent to running
        # the "sphinx-build" command (with the added bonus of running under the
        # same version of Python as the active Python interpreter).
        '-m', 'sphinx',

        # Avoid caching data into a "{OUTPUTDIR}/.doctrees/" subdirectory.
        # Although typically advisable, "{OUTPUTDIR}" is an ignorable temporary
        # test-specific directory deleted immediately after completion of this
        # test. Caching data would only needlessly consume time and space.
        '-E',

        # Enable the HTML mode, rendering HTML-specific documentation files.
        # Although technically arbitrary, this is the typical default mode.
        '-b', 'html',

        # Suffix raised exceptions with tracebacks. This is *CRITICAL.* By
        # default, Sphinx insanely emits *ONLY* useless exception messages for
        # fatal "autodoc" errors resembling:
        #     WARNING: autodoc: failed to import module 'beartype_sphinx'; the
        #     following exception was raised:
        #     cannot import name 'BeartypeConf' from partially initialized
        #     module 'beartype' (most likely due to a circular import)
        '-T',

        # Treat non-fatal warnings as fatal errors. This is *CRITICAL.* By
        # default, Sphinx insanely emits non-fatal warnings for fatal "autodoc"
        # errors resembling:
        #     WARNING: autodoc: failed to import module 'beartype_sphinx'; the
        #     following exception was raised:
        #     No module named 'beartype_sphinx'
        '-W',

        # Absolute or relative dirname of a test-specific subdirectory
        # containing a sample Sphinx structure exercising edge cases in the
        # @beartype decorator.
        str(get_test_func_data_lib_sphinx_dir()),

        # Absolute or relative dirname of a test-specific temporary directory to
        # which Sphinx will emit ignorable rendered documentation files.
        str(tmp_path),
    )

    # Run this command, raising an exception on subprocess failure while
    # forwarding all standard output and error output by this subprocess to the
    # standard output and error file handles of the active Python process.
    run_command_forward_output(command_words=SPHINX_ARGS)


#FIXME: *IGNORE EVERYTHING BELOW.* Fundamentally refactor this integration test
#from Sphinx to MkDocs. Specifically:
#* Define a new "test_mkdocs" submodule.
#* Rename this test to simply test_mkdocs_docs() in that submodule.
#FIXME: For unknown reasons, this currently fails despite working for literally
#years. The culprit is almost certainly the "sphinx.testing.fixtures" API, which
#has yet to be properly documented. That said, we've personally validated that
#this API does still define the expected "make_app" fixture. So, it's unclear
#exactly why our GitHub Actions-based continuous integration (CI) workflow is
#complaining about that fixture being undefined:
#    ___________________ ERROR at setup of test_sphinx_docs_these ___________________
#    file /home/runner/work/beartype/beartype/beartype_test/a90_func/z90_lib/a00_sphinx/test_sphinx.py, line 105
#      @skip_unless_package('sphinx')
#      def test_sphinx_docs_these(make_app, tmp_path) -> None:
#    E       fixture 'make_app' not found
#    >       available fixtures: cache, capfd, capfdbinary, caplog, capsys, capsysbinary, doctest_namespace, monkeypatch, pytestconfig, record_property, record_testsuite_property, record_xml_attribute, recwarn, tmp_path, tmp_path_factory, tmpdir, tmpdir_factory
#    >       use 'pytest --fixtures [testpath]' for help on them.
#    
#    /home/runner/work/beartype/beartype/beartype_test/a90_func/z90_lib/a00_sphinx/test_sphinx.py:105
#FIXME: For the benefit of the community, externally document how to do this
#for others at this open issue:
#    https://github.com/sphinx-doc/sphinx/issues/7008
#Note the trivial "conftest" submodule in this directory. Since this is all
#surprisingly easy, a simple comment describing this should easily suffice.
@skip('Currently broken due to Sphinx breaking backward compatibility.')
@skip_unless_package('sphinx')
def test_sphinx_docs_these(make_app, tmp_path) -> None:
    '''
    Functional test validating that the third-party :mod:`sphinx` package
    successfully builds ReadTheDocs (RTD)-hosted documentation for this project
    residing in the top-level ``doc/`` subdirectory.

    To do so, this test *effectively* externally runs the ``sphinx-build``
    command against our ``doc/src/` Sphinx document tree.

    Parameters
    ----------
    make_app : sphinx.testing.util.SphinxTestApp
        Factory fixture creating and returning a :mod:`pytest`-friendly Sphinx
        object encapsulating the process of running the ``sphinx-build``
        command with the passed parameters.
    tmp_path : pathlib.Path
        Abstract path encapsulating a temporary directory unique to this unit
        test, created in the base temporary directory.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype_test._util.command.pytcmdrun import run_command_forward_output
    from beartype_test._util.path.pytpathmain import get_main_sphinx_source_dir
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)

    # ..................{ SPHINX-BUILD                       }..................
    # Tuple of all shell words with which to run the "sphinx-build" command.
    SPHINX_ARGS = get_interpreter_command_words() + (
        # Fully-qualified name of the Sphinx package to be run. Interestingly,
        # running this package as a script is semantically equivalent to running
        # the "sphinx-build" command (with the added bonus of running under the
        # same version of Python as the active Python interpreter).
        '-m', 'sphinx',

        # Avoid caching data into a "{OUTPUTDIR}/.doctrees/" subdirectory.
        # Although typically advisable, "{OUTPUTDIR}" is an ignorable temporary
        # test-specific directory deleted immediately after completion of this
        # test. Caching data would only needlessly consume time and space.
        '-E',

        # Enable the HTML mode, rendering HTML-specific documentation files.
        # Although technically arbitrary, this is the typical default mode.
        '-b', 'html',

        #FIXME: Reenable once working. Sadly, Sphinx is currently emitting
        #warnings we have *NO* idea how to reasonably resolve. *sigh*
        # # Treat non-fatal warnings as fatal errors. This is *CRITICAL.* By
        # # default, Sphinx insanely emits non-fatal warnings for fatal "autodoc"
        # # errors resembling:
        # #     WARNING: autodoc: failed to import module 'beartype_sphinx'; the
        # #     following exception was raised:
        # #     No module named 'beartype_sphinx'
        # '-W',

        # Absolute or relative dirname of a test-specific subdirectory
        # containing a sample Sphinx structure exercising edge cases in the
        # @beartype decorator.
        str(get_main_sphinx_source_dir()),

        # Absolute or relative dirname of a test-specific temporary directory to
        # which Sphinx will emit ignorable rendered documentation files.
        str(tmp_path),
    )

    # Run this command, raising an exception on subprocess failure while
    # forwarding all standard output and error output by this subprocess to the
    # standard output and error file handles of the active Python process.
    run_command_forward_output(command_words=SPHINX_ARGS)
